"""
The module generates the linear programming relaxation associated to a fanout
or non-fanout inflation instance (see arXiv:1707.06476).

@authors: Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens
"""

from collections import Counter, defaultdict
from functools import reduce, cached_property
from gc import collect
from operator import neg as op_neg
from numbers import Real
from typing import List, Dict, Union, Any
from warnings import warn

import numpy as np
import sympy as sp
from scipy.sparse import coo_array, vstack
from tqdm import tqdm

from .. import InflationProblem
from .lp_utils import solveLP
from .monomial_classes import InternalAtomicMonomial, CompoundMoment
from .numbafied import (nb_outer_bitwise_or,
                        nb_outer_bitwise_xor)
from .writer_utils import write_to_lp, write_to_mps
from ..sdp.fast_npa import nb_is_knowable as is_knowable
from ..sdp.quantum_tools import flatten_symbolic_powers
from ..utils import clean_coefficients, eprint, partsextractor, \
    expand_sparse_vec


class InflationLP(object):
    constant_term_name = "constant_term"

    def __init__(self,
                 inflationproblem: InflationProblem,
                 nonfanout: bool = True,
                 local_level: int = None,
                 supports_problem: bool = False,
                 default_non_negative: bool = True,
                 include_all_outcomes: bool = False,
                 verbose: int = None) -> None:
        """
        Class for generating and solving an LP relaxation for fanout or 
        nonfanout inflation.

        Parameters
        ----------
        inflationproblem : InflationProblem
            Details of the scenario.
        nonfanout : bool, optional
            Whether to consider only nonfanout inflation (GPT problem) or
            fanout inflation (classical problem). By default ``False``.
        local_level : int, optional
            If specified, do not assume nonnegative probability distribution
             over all variables, but only over marginal sets of limited size.
        supports_problem : bool, optional
            Whether to consider feasibility problems with distributions, or 
            just with the distribution's support. By default ``False``.
        default_non_negative : bool, optional
            Whether to set all variables to be non-negative by default. By
            default ``True``.
        include_all_outcomes : bool, optional
            Whether to include all outcomes in the LP, as opposed to using
            Collins-Gisin notation when possible. By default ``False``.
        verbose : int, optional
            Optional parameter for level of verbose:
                * 0: quiet (default),
                * 1: monitor level: track program process and show warnings,
                * 2: debug level: show properties of objects created.
        """
        self.problem_type = "lp"
        self.supports_problem = supports_problem
        if verbose is not None:
            if inflationproblem.verbose > verbose:
                warn("Overriding the verbosity from InflationProblem")
            self.verbose = verbose
        else:
            self.verbose = inflationproblem.verbose
        self.local_level = local_level
        if not nonfanout:
            assert not inflationproblem._default_notcomm.any(), \
                "You appear to be requesting fanout (classical)" \
                    + " inflation, \nbut have not specified classical_sources=`all`." \
                    + "\nNote that the `nonfanout` keyword argument is deprecated as of release 2.0.0"

        self.all_operators_commute = True
        self.all_commuting_q_2d = lambda mon: True
        self.all_commuting_q_1d = lambda lexmon: True

        self.default_non_negative = default_non_negative
        if self.verbose > 1:
            print(inflationproblem)

        # Inheritance
        self.InflationProblem = inflationproblem
        self.names = inflationproblem.names
        self.names_to_ints = inflationproblem.names_to_ints
        self._lexrepr_to_names = inflationproblem._lexrepr_to_names
        self._lexrepr_to_copy_index_free_names = inflationproblem._lexrepr_to_copy_index_free_names
        self.op_from_name = {}
        for i, op_names in enumerate(inflationproblem._lexrepr_to_all_names.tolist()):
            for op_name in op_names:
                self.op_from_name.setdefault(op_name, i)
        self.nr_sources = inflationproblem.nr_sources
        self.nr_parties = inflationproblem.nr_parties
        self.hypergraph = inflationproblem.hypergraph
        self.network_scenario = inflationproblem.is_network
        self.inflation_levels = inflationproblem.inflation_level_per_source
        self.setting_cardinalities = inflationproblem.settings_per_party
        self.outcome_cardinalities = inflationproblem.outcomes_per_party
        self.private_setting_cardinalities = inflationproblem.private_settings_per_party
        self.expected_distro_shape = inflationproblem.expected_distro_shape
        self.rectify_fake_setting = inflationproblem.rectify_fake_setting
        self.factorize_monomial_1d = inflationproblem.factorize_monomial_1d
        self._is_knowable_q_non_networks = \
            inflationproblem._is_knowable_q_non_networks
        self._nr_properties = inflationproblem._nr_properties
        self.np_dtype = inflationproblem._np_dtype
        self._lexorder = inflationproblem._lexorder
        self._nr_operators = inflationproblem._nr_operators
        self._lexeye = np.eye(self._nr_operators, dtype=bool)
        self._lexrange = np.arange(self._nr_operators)
        self.lexorder_symmetries = inflationproblem.symmetries
        self._from_2dndarray = inflationproblem._from_2dndarray
        self.mon_to_lexrepr = inflationproblem.mon_to_lexrepr
        self.blank_bool_vec = np.zeros(self._nr_operators, dtype=bool)
        self.blank_bool_array = np.zeros((1, self._nr_operators), dtype=bool)
        self._ortho_groups_per_party = inflationproblem._ortho_groups_per_party
        self.has_children = inflationproblem.has_children.copy()
        if include_all_outcomes or supports_problem: # HACK to fix detection of incompatible supports. (Can be fixed upon adding set_extra_equalities)
            self.has_children[:] = True
        self.does_not_have_children = np.logical_not(self.has_children)

        all_ortho_groups_as_boolarrays = []
        for ortho_groups in inflationproblem._ortho_idxs_per_party:
            ortho_groups_as_boolarrays = []
            for ortho_group in ortho_groups:
                ortho_groups_as_boolarrays.append(self._lexeye[ortho_group])
            all_ortho_groups_as_boolarrays.append(ortho_groups_as_boolarrays)
        self._CG_adjusting_ortho_groups_as_boolarrays = []
        self._CG_nonadjusting_ortho_groups_as_boolarrays = []
        self._all_ortho_groups_as_boolarrays = []
        self._CG_limited_ortho_groups_as_boolarrays = []
        for i, ortho_groups_as_boolarrays in enumerate(all_ortho_groups_as_boolarrays):
            if self.does_not_have_children[i]:
                self._CG_adjusting_ortho_groups_as_boolarrays.extend(ortho_groups_as_boolarrays)
                for ortho_group_as_boolarray in ortho_groups_as_boolarrays:
                    self._CG_limited_ortho_groups_as_boolarrays.append(ortho_group_as_boolarray[:-1])
            else:
                self._CG_nonadjusting_ortho_groups_as_boolarrays.extend(ortho_groups_as_boolarrays)
                self._CG_limited_ortho_groups_as_boolarrays.extend(ortho_groups_as_boolarrays)
            self._all_ortho_groups_as_boolarrays.extend(ortho_groups_as_boolarrays)

        # We want to consider ALL outcomes for variables which have children, 
        # but not the last outcome for childless variables.
        # In other words, if we take all the CG equalities, we want to convert 
        # them to inequalities when the LHS is non_CG but childless-only.
        if np.any(self.does_not_have_children):
            bad_boolvecs_for_ineqs = [bool_array[-1] for bool_array in self._CG_adjusting_ortho_groups_as_boolarrays]
            self._boolvec_for_CG_ineqs = np.bitwise_or.reduce(bad_boolvecs_for_ineqs, axis=0)
        else:
            self._boolvec_for_CG_ineqs = self.blank_bool_vec
        if np.any(self.has_children):
            bad_boolvecs_for_eqs = [bool_array[0] for bool_array in self._CG_nonadjusting_ortho_groups_as_boolarrays]
            self._boolvec_for_FR_eqs = np.bitwise_or.reduce(bad_boolvecs_for_eqs, axis=0)
        else:
            self._boolvec_for_FR_eqs = self.blank_bool_vec

        if self.verbose > 1:
            eprint("Number of single operator measurements per party:", end="")
            prefix = " "
            for i, measures in enumerate(inflationproblem.measurements):
                op_count = np.prod(measures.shape[:2])
                eprint(prefix + f"{self.names[i]}={op_count}", end="")
                prefix = ", "
            eprint()
        self.use_lpi_constraints = False

        self.identity_operator = np.empty((0, self._nr_properties),
                                          dtype=self.np_dtype)
        self.zero_operator = np.zeros((1, self._nr_properties),
                                      dtype=self.np_dtype)

        # These next properties are reset during generate_lp, but are needed in
        # init so as to be able to test the Monomial constructor function
        # without generate_lp.

        self._atomic_monomial_from_hash  = {}
        self.monomial_from_atoms        = {}
        self.monomial_from_name = {}
        self.monomial_from_symbol = {}
        self.One  = self.Monomial(self.identity_operator, idx=1)
        self._generate_lp()

    ###########################################################################
    # MAIN ROUTINES EXPOSED TO THE USER                                       #
    ###########################################################################
    @cached_property
    def monomials_as_strings(self):
        """Returns the monomials as strings."""
        return [mon.name for mon in self.monomials]

    def set_bounds(self,
                   bounds: Union[dict, None],
                   bound_type: str = "up") -> None:
        r"""Set numerical lower or upper bounds on the moments generated in the
        LP relaxation. The bounds are at the level of the LP variables,
        and do not take into consideration non-convex constraints. E.g., two
        individual lower bounds, :math:`p_A(0|0) \geq 0.1` and
        :math:`p_B(0|0) \geq 0.1` do not directly impose the constraint
        :math:`p_A(0|0)*p_B(0|0) \geq 0.01`, which should be set manually if
        needed.

        Parameters
        ----------
        bounds : Union[dict, None]
            A dictionary with keys as monomials and values being the bounds.
            The keys can be either CompoundMonomial objects, or names (`str`)
            of Monomial objects.
        bound_type : str, optional
            Specifies whether we are setting upper (``"up"``) or lower
            (``"lo"``) bounds, by default "up".

        Examples
        --------
        >>> self.set_bounds({"pAB(00|00)": 0.2}, "lo")
        """
        assert bound_type in ["up", "lo"], \
            "The 'bound_type' argument should be either 'up' or 'lo'"
        if bound_type == "up":
            self._set_upperbounds(bounds)
        else:
            self._set_lowerbounds(bounds)

    @cached_property
    def atomic_factors(self):
        atoms = set()
        for mon in self.monomials:
            atoms.update(mon.factors)
        return sorted(atoms)

    @cached_property
    def atomic_monomials(self):
        """Returns the atomic monomials."""
        return [self._monomial_from_atoms([atom]) for atom in
                self.atomic_factors]

    @cached_property
    def atom_from_name(self):
        """Returns the atomic monomials."""
        lookup_dict = {}
        for atom in self.atomic_factors:
            lookup_dict[atom.name] = atom
            lookup_dict[atom.legacy_name] = atom
        return lookup_dict

    @cached_property
    def knowable_atoms(self):
        """Returns the knowable atoms."""
        return [m for m in self.atomic_monomials if m.is_knowable]

    @cached_property
    def do_conditional_atoms(self):
        """Returns the atomic monomials which correspond to do conditionals."""
        return [m for m in self.atomic_monomials if
                (m.is_do_conditional and not m.is_knowable)]

    @cached_property
    def factorization_conditions(self):
        """Returns the factorization conditions."""
        conds = {}
        for mon in self.monomials:
            if mon.n_factors > 1:
                conds[mon] = tuple(self.monomial_from_atoms[(fac,)]
                                   for fac in mon.factors)
        return conds

    @cached_property
    def quadratic_factorization_conditions(self):
        """Returns the quadratic factorization conditions."""
        conds = {}
        for mon in self.monomials:
            if mon.n_factors > 1:
                conds[mon] = (self.monomial_from_atoms[mon.factors[:1]],
                              self.monomial_from_atoms[mon.factors[1:]])
        return conds

    def set_distribution(self,
                         prob_array: Union[np.ndarray, None],
                         use_lpi_constraints: bool = False,
                         shared_randomness: bool = False) -> None:
        r"""Set numerically all the knowable (and optionally semiknowable)
        moments according to the probability distribution specified.

        Parameters
        ----------
        prob_array : numpy.ndarray
            Multidimensional array encoding the distribution, which is
            called as ``prob_array[a,b,c,...,x,y,z,...]`` where
            :math:`a,b,c,\dots` are outputs and :math:`x,y,z,\dots` are
            inputs. Note: even if the inputs have cardinality 1 they must
            be specified, and the corresponding axis dimensions are 1.
            The parties' outcomes and measurements must be appear in the
            same order as specified by the ``order`` parameter in the
            ``InflationProblem`` used to instantiate ``InflationLP``.
        use_lpi_constraints : bool, optional
            Specification whether linearized polynomial constraints (see,
            e.g., Eq. (D6) in `arXiv:2203.16543
            <https://www.arxiv.org/abs/2203.16543/>`_) will be imposed or
            not. By default ``False``.
        shared_randomness : bool, optional
            Specification whether higher order monomials may be calculated.
            If universal shared randomness is present (i.e., the flag is
            ``True``), only atomic monomials are assigned numerical values.
        """
        if prob_array is not None:
            assert prob_array.shape == self.expected_distro_shape, \
                f"Cardinalities mismatch: \n" \
                f"expected {self.expected_distro_shape}, \n " \
                f"got {prob_array.shape}"
            knowable_values = {atom: atom.compute_marginal(prob_array)
                               for atom in self.knowable_atoms}
        else:
            knowable_values = {}

        self.set_values(knowable_values,
                        use_lpi_constraints=use_lpi_constraints,
                        only_specified_values=shared_randomness)

    def set_objective(self,
                      objective: Union[sp.core.expr.Expr, dict, None],
                      direction: str = "max") -> None:
        """Set or change the objective function of the polynomial optimization
        problem.

        Parameters
        ----------
        objective : Union[sp.core.expr.Expr, dict, None]
            The objective function, either as a combination of sympy symbols,
            as a dictionary with keys the monomials or their names, and as
            values the corresponding coefficients, or ``None`` for clearing
            a previous objective.
        direction : str, optional
            Direction of the optimization (``"max"``/``"min"``). By default
            ``"max"``.
        """
        assert (not self.supports_problem) or (objective is None), \
            "Supports problems do not support specifying objective functions."
        assert direction in ["max", "min"], ("The 'direction' argument should "
                                             + " be either 'max' or 'min'")

        self._reset_objective()
        if direction == "max":
            self.maximize = True
        else:
            self.maximize = False
        if objective is None:
            return
        else:
            if self.use_lpi_constraints and self.verbose > 0:
                warn("You have the flag `use_lpi_constraints` set to True. Be "
                     + "aware that imposing linearized polynomial constraints "
                     + "will constrain the optimization to distributions with "
                     + "fixed marginals.")
            sign = (1 if self.maximize else -1)
            self.objective = {mon: (sign * coeff)
                              for mon, coeff in self._sanitise_dict(objective).items()}
            surprising_objective_terms = {mon for mon in self.objective.keys()
                                          if mon not in self.monomials}
            assert len(surprising_objective_terms) == 0, \
                ("When interpreting the objective we have encountered at " +
                 "least one monomial that does not appear in the original " +
                 f"generating set:\n\t{surprising_objective_terms}")

    def update_values(self,
                      values: Union[Dict[Union[CompoundMoment,
                      InternalAtomicMonomial,
                      sp.core.symbol.Symbol,
                      str],
                      Union[float, sp.core.expr.Expr]],
                      None],
                      use_lpi_constraints: bool = False,
                      only_specified_values: bool = False) -> None:
        """Directly assign numerical values to variables in the generating set.
        This is done via a dictionary where keys are the variables to have
        numerical values assigned (either in their operator form, in string
        form, or directly referring to the variable in the generating set), and
        the values are the corresponding numerical quantities.

        Parameters
        ----------
        values : Union[None, Dict[Union[CompoundMoment, InternalAtomicMonomial, sympy.core.symbol.Symbol, str], float]]
            The description of the variables to be assigned numerical values
            and the corresponding values. The keys can be either of the
            Monomial class, symbols or strings (which should be the name of
            some Monomial).
        use_lpi_constraints : bool
            Specification whether linearized polynomial constraints (see, e.g.,
            Eq. (D6) in `arXiv:2203.16543
            <https://www.arxiv.org/abs/2203.16543/>`_) will be imposed or not.
            By default ``False``.
        only_specified_values : bool
            Specifies whether one wishes to fix only the variables provided
            (``True``), or also the variables containing products of the
            monomials fixed (``False``). Regardless of this flag, unknowable
            variables can also be fixed.
        """
        if (values is None) or len(values) == 0:
            return

        self._reset_solution()

        self.use_lpi_constraints = use_lpi_constraints

        if (len(self.objective) > 1) and self.use_lpi_constraints:
            warn("You have an objective function set. Be aware that imposing "
                 + "linearized polynomial constraints will constrain the "
                 + "optimization to distributions with fixed marginals.")
        for mon, value in values.items():
            try:  # We do this to deal with NaNs (such as in undefined supports)
                  # in a way that is also compatible with symbolic values.
                if np.isnan(value):
                    continue
            except TypeError:
                pass
            mon = self._sanitise_monomial(mon)
            self.known_moments[mon] = value
        if not only_specified_values:
            atomic_knowns = {mon.factors[0]: val
                             for mon, val in self.known_moments.items()
                             if len(mon) == 1}
            monomials_not_present = set(self.known_moments.keys()
                                        ).difference(self.monomials)
            for mon in monomials_not_present:
                warn(f"We do not recognize the set value of {mon} "
                     + "as it does not appear in the internal list of monomials.")
                del self.known_moments[mon]

            # Get the remaining monomials that need assignment
            if all(atom.is_knowable for atom in atomic_knowns):
                if not self.use_lpi_constraints:
                    remaining_mons = (mon for mon in self.monomials
                                      if ((not mon.is_atomic)
                                          and mon.is_knowable))
                else:
                    remaining_mons = (mon for mon in self.monomials
                                      if ((not mon.is_atomic)
                                          and mon.knowability_status
                                          in ["Knowable", "Semi"]))
            else:
                remaining_mons = (mon for mon in self.monomials
                                  if not mon.is_atomic)
            surprising_semiknowns = set()
            for mon in tqdm(list(remaining_mons), disable=not self.verbose,
                                 desc="Evaluating monomials    "):
                value, unknown_factors, known_status = mon.evaluate(
                    atomic_knowns,
                    self.use_lpi_constraints)
                if known_status == "Known":
                    self.known_moments[mon] = value
                elif known_status == "Semi":
                    if self.use_lpi_constraints:
                        unknown_mon = \
                            self._monomial_from_atoms(unknown_factors)
                        self.semiknown_moments[mon] = (value, unknown_mon)
                        if self.verbose > 0:
                            if unknown_mon not in self.monomials:
                                surprising_semiknowns.add(unknown_mon)
                                raise Exception(
                                    "When processing LPI constraints we encountered a " +
                                    "monomial that does not appear in the list of LP " +
                                    f"variables:\n\t{unknown_mon}")
                else:
                    pass
            if (len(surprising_semiknowns) >= 1) and (self.verbose > 0):
                warn("When processing LPI constraints we encountered at " +
                     "least one monomial that does not appear in the " +
                     f"generating set:\n\t{surprising_semiknowns}")
            del atomic_knowns, surprising_semiknowns
        self._cleanup_after_set_values()

    def set_values(self, 
                   values: Union[Dict[Union[CompoundMoment,
                   InternalAtomicMonomial,
                   sp.core.symbol.Symbol,
                   str],
                   Union[float, sp.core.expr.Expr]],
                   None], 
                   **kwargs):
        r"""Exactly like update_values, except it resets all known values to
        zero as an intermediate step.
        
        Parameters
        ----------
        values : Union[None, Dict[Union[CompoundMoment, InternalAtomicMonomial, sympy.core.symbol.Symbol, str], float]]
            The description of the variables to be assigned numerical values
            and the corresponding values. The keys can be either of the
            Monomial class, symbols or strings (which should be the name of
            some Monomial).
        """
        self._reset_values()
        if (values is None) or len(values) == 0:
            self._cleanup_after_set_values()
            return
        else:
            self.update_values(values, **kwargs)
            return

    def set_extra_equalities(self,
                             extra_equalities: Union[list, None]) -> None:
        """Set extra equality constraints for the LP.

        Parameters
        ----------
        extra_equalities : Union[list, None]
            List of dictionaries representing additional equality constraints.
            The keys (variables) can be strings, instances of
            `CompoundMonomial`, or monomials as 2D arrays.
        """
        self.extra_equalities = []  # Reset every time
        if not extra_equalities or extra_equalities is None:
            return
        self.extra_equalities = [self._sanitise_dict(eq)
                                 for eq in extra_equalities]

    def set_extra_inequalities(self,
                               extra_inequalities: Union[list, None]) -> None:
        """Set extra inequality constraints for the LP.

        Parameters
        ----------
        extra_inequalities : Union[list, None]
            List of dictionaries representing additional inequality
             constraints. The keys (variables) can be strings, instances of
            `CompoundMonomial`, or monomials as 2D arrays.
        """
        self.extra_inequalities = []  # Reset every time
        if not extra_inequalities or extra_inequalities is None:
            return
        self.extra_inequalities = [self._sanitise_dict(ineq)
                                   for ineq in extra_inequalities]

    def solve(self,
              relax_known_vars: bool = False,
              relax_inequalities: bool = False,
              solve_dual: bool = True,
              solverparameters: dict = None,
              verbose: int = None,
              default_non_negative: bool = None,
              **solver_arguments) -> None:
        r"""Call a solver on the LP relaxation. Upon successful solution, it
        returns the primal and dual objective values along with the solution
        matrices.

        Parameters
        ----------
        relax_known_vars : bool, optional
            Do feasibility as optimization where each known value equality
            becomes two relaxed inequality constraints. E.g., P(A) = 0.7
            becomes P(A) + lambda >= 0.7 and P(A) - lambda <= 0.7, where lambda
            is a slack variable. By default, ``False``.
        relax_inequalities : bool, optional
            Do feasibility as optimization where each inequality is relaxed by
            the non-negative slack variable lambda. By default, ``False``.
        solve_dual : bool, optional
            Optimize the dual problem (recommended). By default ``False``.
        solverparameters : dict, optional
            Extra parameters to be sent to the solver. By default ``None``.
        verbose : int, optional
            Verbosity level.
        default_non_negative : bool, optional
            If ``True``, all variables are set to be non-negative by default.
        solver_arguments : dict, optional
            By default, solve will use the dictionary of LP keyword arguments
            given by ``_prepare_solver_arguments()``. However, a user may
            manually override these arguments by passing their own here.
        """
        if (relax_known_vars or relax_inequalities) and \
                len(self.objective) > 0:
            warn("You have a non-trivial objective, but set to solve a "
                 "feasibility problem as optimization. Setting "
                 "relax_known_vars=False, relax_inequalities=False, and "
                 "optimizing the objective...")
            relax_known_vars = relax_inequalities = False
        if verbose is None:
            real_verbose = self.verbose
        else:
            real_verbose = verbose
        if default_non_negative is None:
            real_default_non_negative = self.default_non_negative
        else:
            real_default_non_negative = default_non_negative

        args = self._prepare_solver_matrices()
        
        # Still allow for 'feas_as_optim' as an argument
        if 'feas_as_optim' in solver_arguments:
            if solver_arguments["feas_as_optim"]:
                relax_known_vars = False
                relax_inequalities = True
                del solver_arguments["feas_as_optim"]
            else:
                relax_known_vars = relax_inequalities = False
                del solver_arguments["feas_as_optim"]
        args.update(solver_arguments)

        args.update({"relax_known_vars": relax_known_vars,
                     "relax_inequalities": relax_inequalities,
                     "verbose": real_verbose,
                     "default_non_negative": real_default_non_negative,
                     "solverparameters": solverparameters,
                     "solve_dual": solve_dual})

        self.solution_object = solveLP(**args)
        self.success = self.solution_object["success"]
        self.status = self.solution_object["status"]
        if self.success:
            self.primal_objective = self.solution_object["primal_value"]
            self.objective_value  = self.solution_object["primal_value"]
            self.objective_value *= (1 if self.maximize else -1)
        else:
            self.primal_objective = self.status
            self.objective_value  = self.status
        collect()

    ###########################################################################
    # PUBLIC ROUTINES RELATED TO THE PROCESSING OF CERTIFICATES               #
    ###########################################################################
    def certificate_as_dict(self,
                             clean: bool = True,
                             chop_tol: float = 1e-10,
                             round_decimals: int = 3) -> dict:
        """Give certificate as dictionary with monomials as keys and
        their coefficients in the certificate as the values. The certificate
        of incompatibility is ``cert < 0``.

        If the certificate is evaluated on a point giving a negative value, this
        guarantees that the compatibility test for the same point is infeasible
        provided the set of constraints of the program does not change. Warning:
        when using ``use_lpi_constraints=True`` the set of constraints depends
        on the specified distribution, thus the certificate is not guaranteed to
        apply.

        Parameters
        ----------
        clean : bool, optional
            If ``True``, eliminate all coefficients that are smaller than
            ``chop_tol``, normalise and round to the number of decimals
            specified by ``round_decimals``. By default ``True``.
        chop_tol : float, optional
            Coefficients in the dual certificate smaller in absolute value are
            set to zero. By default ``1e-10``.
        round_decimals : int, optional
            Coefficients that are not set to zero are rounded to the number of
            decimals specified. By default ``3``.

        Returns
        -------
        dict
            The expression of the certificate in terms of probabilities and
            marginals. The certificate of incompatibility is ``cert < 0``.
        """
        try:
            dual = self.solution_object["dual_certificate"]
        except AttributeError:
            raise Exception("For extracting a certificate you need to solve " +
                            "a problem. Call \"InflationLP.solve()\" first.")
        if len(self.semiknown_moments) > 0:
            warn("Beware that, because the problem contains linearized " +
                 "polynomial constraints, the certificate is not guaranteed " +
                 "to apply to other distributions.")
        if np.allclose(list(dual.values()), 0.):
            return {}
        if clean:
            dual = clean_coefficients(dual, chop_tol, round_decimals)
        return {self.monomial_from_name[k]: v for k, v in dual.items()
                if not self.monomial_from_name[k].is_zero}

    def probs_from_dict(self,
                        dict_with_monomial_keys: dict) -> sp.core.add.Add:
        """Converts a monomial dictionary into a SymPy expression.

        Parameters
        ----------
        dict_with_monomial_keys : Dict[sympy.Symbol, float]
            Dictionary with monomials and associated coefficients.

        Returns
        -------
        sympy.core.add.Add
            The expression of the polynomial encoded in the dictionary.
        """
        polynomial = sp.S.Zero
        for mon, coeff in self._sanitise_dict(dict_with_monomial_keys).items():
            polynomial += coeff * mon.symbol
        return polynomial

    def certificate_as_probs(self,
                             clean: bool = True,
                             chop_tol: float = 1e-10,
                             round_decimals: int = 3) -> sp.core.add.Add:
        """Give certificate as symbolic sum of probabilities. The certificate
        of incompatibility is ``cert < 0``.

        If the certificate is evaluated on a point giving a negative value, this
        guarantees that the compatibility test for the same point is infeasible
        provided the set of constraints of the program does not change. Warning:
        when using ``use_lpi_constraints=True`` the set of constraints depends
        on the specified distribution, thus the certificate is not guaranteed to
        apply.

        Parameters
        ----------
        clean : bool, optional
            If ``True``, eliminate all coefficients that are smaller than
            ``chop_tol``, normalise and round to the number of decimals
            specified by ``round_decimals``. By default ``True``.
        chop_tol : float, optional
            Coefficients in the dual certificate smaller in absolute value are
            set to zero. By default ``1e-10``.
        round_decimals : int, optional
            Coefficients that are not set to zero are rounded to the number of
            decimals specified. By default ``3``.

        Returns
        -------
        sympy.core.add.Add
            The expression of the certificate in terms of probabilities and
            marginals. The certificate of incompatibility is ``cert < 0``.
        """

        return self.probs_from_dict(self.certificate_as_dict(
                clean=clean,
                chop_tol=chop_tol,
                round_decimals=round_decimals))

    def string_from_dict(self,
                         dict_with_monomial_keys) -> str:
        """Converts a monomial dictionary into a string.

        Parameters
        ----------
        dict_with_monomial_keys : Dict[sympy.Symbol, float]
            Dictionary with monomials and associated coefficients.

        Returns
        -------
        str
            The expression of the certificate in string form.
        """
        as_dict = self._sanitise_dict(dict_with_monomial_keys)
        # Watch out for when "1" is note the same as "constant_term"
        constant_value = as_dict.pop(self.Constant_Term,
                                     as_dict.pop(self.One, 0.)
                                     )
        if constant_value:
            polynomial_as_str = str(constant_value)
        else:
            polynomial_as_str = ""
        for mon, coeff in as_dict.items():
            if mon.is_zero or np.isclose(np.abs(coeff), 0):
                continue
            else:
                polynomial_as_str += "+" if coeff >= 0 else "-"
                if np.isclose(abs(coeff), 1):
                    polynomial_as_str += mon.name
                else:
                    polynomial_as_str += "{0}*{1}".format(abs(coeff), mon.name)
        if not len(polynomial_as_str):  # Failsafe for empty dictionary.
            polynomial_as_str = "0"
        elif polynomial_as_str[0] == "+":
            polynomial_as_str = polynomial_as_str[1:]
        return polynomial_as_str

    def certificate_as_string(self,
                              clean: bool = True,
                              chop_tol: float = 1e-10,
                              round_decimals: int = 3) -> str:
        """Give the certificate as a string of a sum of probabilities. The
        expression is in the form such that its satisfaction implies
        incompatibility.

        If the certificate is evaluated on a point giving a negative value, this
        guarantees that the compatibility test for the same point is infeasible
        provided the set of constraints of the program does not change. Warning:
        when using ``use_lpi_constraints=True`` the set of constraints depends
        on the specified distribution, thus the certificate is not guaranteed to
        apply.

        Parameters
        ----------
        clean : bool, optional
            If ``True``, eliminate all coefficients that are smaller than
            ``chop_tol``, normalise and round to the number of decimals
            specified by ``round_decimals``. By default, ``True``.
        chop_tol : float, optional
            Coefficients in the dual certificate smaller in absolute value are
            set to zero. By default, ``1e-10``.
        round_decimals : int, optional
            Coefficients that are not set to zero are rounded to the number of
            decimals specified. By default, ``3``.

        Returns
        -------
        str
            The certificate in terms of probabilities and marginals. The
            certificate of incompatibility is ``cert < 0``.
        """
        return self.string_from_dict(
            self.certificate_as_dict(
                clean=clean,
                chop_tol=chop_tol,
                round_decimals=round_decimals)) + " < 0"

    def evaluate_polynomial(self, polynomial: dict, prob_array: np.ndarray):
        """Evaluate the certificate of infeasibility in a target probability
        distribution. If the evaluation is a negative value, the distribution is
        not compatible with the causal structure. Warning: when using
        ``use_lpi_constraints=True`` the set of constraints depends on the
        specified distribution, thus the certificate is not guaranteed to apply.

        Parameters
        ----------
        polynomial : dict
            A dictionary of monomials with coefficients as values.
        prob_array : numpy.ndarray
            Multidimensional array encoding the distribution, which is
            called as ``prob_array[a,b,c,...,x,y,z,...]`` where
            :math:`a,b,c,\\dots` are outputs and :math:`x,y,z,\\dots` are
            inputs. Note: even if the inputs have cardinality 1 they must
            be specified, and the corresponding axis dimensions are 1.
            The parties' outcomes and measurements must appear in the
            same order as specified by the ``order`` parameter in the
            ``InflationProblem`` used to instantiate ``InflationLP``.

        Returns
        -------
        float
            The evaluation of the certificate of infeasibility in prob_array.
        """
        return sum((atom.compute_marginal(prob_array) * val
                    for atom, val in self._sanitise_dict(polynomial).items()))


    def evaluate_certificate(self, prob_array: np.ndarray) -> float:
        """Evaluate the certificate of infeasibility in a target probability
        distribution. If the evaluation is a negative value, the distribution is
        not compatible with the causal structure. Warning: when using
        ``use_lpi_constraints=True`` the set of constraints depends on the
        specified distribution, thus the certificate is not guaranteed to apply.

        Parameters
        ----------
        prob_array : numpy.ndarray
            Multidimensional array encoding the distribution, which is
            called as ``prob_array[a,b,c,...,x,y,z,...]`` where
            :math:`a,b,c,\\dots` are outputs and :math:`x,y,z,\\dots` are
            inputs. Note: even if the inputs have cardinality 1 they must
            be specified, and the corresponding axis dimensions are 1.
            The parties' outcomes and measurements must appear in the
            same order as specified by the ``order`` parameter in the
            ``InflationProblem`` used to instantiate ``InflationLP``.

        Returns
        -------
        float
            The evaluation of the certificate of infeasibility in prob_array.
        """
        if self.use_lpi_constraints:
            warn("You have used LPI constraints to obtain the certificate. " +
                 "Be aware that, because of that, the certificate may not be " +
                 "valid for other distributions.")
        return self.evaluate_polynomial(self.certificate_as_dict(), prob_array)

    def desymmetrize_certificate(self) -> dict:
        """If the scenario contains symmetries other than the inflation
        symmetries, this function writes a certificate of infeasibility valid
        for non-symmetric distributions too.

        Parameters
        ----------
        replace : bool, optional
            Whether to replace the certificate in ``self.solution_object``.
            By default ``False``.

        Returns
        -------
        dict
            The expression of the un-symmetrized certificate in terms of
            probabilities and marginals. The certificate of incompatibility is
            ``cert < 0``.
        """
        try:
            dual = self.solution_object["dual_certificate"]
        except AttributeError:
            raise Exception("For extracting a certificate you need to solve " +
                            "a problem. Call \"InflationSDP.solve()\" first.")

        desymmetrized = defaultdict(int)
        norm = len(self.InflationProblem.symmetries)
        lexmon_names = self.InflationProblem._lexrepr_to_copy_index_free_names
        for symm in self.InflationProblem.symmetries:
            for mon, coeff in dual.items():
                mon = self.monomial_from_name[mon]
                if not mon.is_zero:
                    desymm_mon = lexmon_names[symm[mon.as_lexmon]]
                    desymm_mon = sorted(desymm_mon, key=lambda x: x[0])
                    if not mon.is_one:
                        desymm_name = "P[" + " ".join(desymm_mon) + "]"
                    else:
                        desymm_name = "1"
                    desymmetrized[desymm_name] += coeff / norm
        return {k:v for k,v in desymmetrized.items()}

    ###########################################################################
    # OTHER ROUTINES EXPOSED TO THE USER                                      #
    ##########################################################################
    def reset(self, which: Union[str, List[str]] = "all") -> None:
        """Reset the various user-specifiable objects in the inflation LP.

        Parameters
        ----------
        which : Union[str, List[str]]
            The objects to be reset. It can be fed as a single string or a list
            of them. Options include ``"bounds"``, ``"lowerbounds"``,
            ``"upperbounds"``, ``"objective"``, ``"values"``, and ``"all"``.
        """
        if type(which) == str:
            if which == "all":
                self.reset(["bounds", "objective", "values"])
            elif which == "bounds":
                self._reset_bounds()
            elif which == "lowerbounds":
                self._reset_lowerbounds()
            elif which == "upperbounds":
                self._reset_upperbounds()
            elif which == "objective":
                self._reset_objective()
            elif which == "values":
                self._reset_values()
            else:
                raise Exception(f"The attribute {which} is not part of " +
                                "InflationLP.")
        else:
            for attr in which:
                self.reset(attr)
        collect()

    def write_to_file(self, filename: str) -> None:
        """Exports the problem to a file.

        Parameters
        ----------
        filename : str
            Name of the exported file. If no file format is specified, it
            defaults to the LP format. Supported formats are ``.lp`` (LP) and
            ``.mps`` (MPS).
        """
        # Determine file extension
        parts = filename.split(".")
        if len(parts) >= 2:
            extension = parts[-1]
        else:
            extension = "lp"
            filename += ".lp"

        # Write file according to the extension
        args = self._prepare_solver_arguments(separate_bounds=True)
        if self.verbose > 0:
            eprint("Writing the LP program to", filename)
        if extension == "lp":
            write_to_lp(args, filename)
        elif extension == "mps":
            write_to_mps(args, filename)
        else:
            raise Exception("File format not supported. Please choose between "
                            + "the extensions `.lp` and `.mps`.")

    ###########################################################################
    # ROUTINES RELATED TO CONSTRUCTING COMPOUND MONOMIAL INSTANCES            #
    ###########################################################################
    def _AtomicMonomial(self,
                        array1d: np.ndarray) -> InternalAtomicMonomial:
        """Construct an instance of the `InternalAtomicMonomial` class from
        a 2D array description of a monomial.

        See the documentation of the `InternalAtomicMonomial` class for more
        details.

        Parameters
        ----------
        array1d : numpy.ndarray
            Monomial encoded as a 1D array of integers relative to _lexorder.

        Returns
        -------
        InternalAtomicMonomial
            An instance of the `InternalAtomicMonomial` class representing the
            input.
        """
        key = self.blank_bool_vec.copy()  # Quantum case will be different
        key[array1d] = True
        try:
            return self._atomic_monomial_from_hash[key.tobytes()]
        except KeyError:
            if len(self.lexorder_symmetries) == 1:
                mon = InternalAtomicMonomial(self, array1d)
                self._atomic_monomial_from_hash[key.tobytes()] = mon
                return mon
            else:
                mon_as_symboolvec = key[self.lexorder_symmetries]
                mon_as_symboolvec = mon_as_symboolvec[
                    np.lexsort(mon_as_symboolvec.T[::-1])]
                mon_as_boolvec = mon_as_symboolvec[-1]
                mon = InternalAtomicMonomial(self, 
                                             np.flatnonzero(mon_as_boolvec))
                for alt_key in mon_as_symboolvec:
                    self._atomic_monomial_from_hash[alt_key.tobytes()] = mon
                return mon

    def Monomial(self, array1d: np.ndarray, idx: int = -1) -> CompoundMoment:
        r"""Create an instance of the `CompoundMonomial` class from a 2D array.
        An instance of `CompoundMonomial` is a collection of
        `InternalAtomicMonomial`.

        Parameters
        ----------
        array1d : numpy.ndarray
            Moment encoded as a 1D array of integers, relative to _lexorder
        idx : int, optional
            Assigns an integer index to the resulting monomial, which can be
            used as an id, by default -1.

        Returns
        -------
        CompoundMoment
            The monomial factorised into AtomicMonomials, all brought to
            representative form under inflation symmetries.
        """
        _factors = self.factorize_monomial_1d(array1d, canonical_order=False)
        list_of_atoms = [self._AtomicMonomial(factor)
                         for factor in _factors if len(factor)]
        mon = self._monomial_from_atoms(list_of_atoms)
        mon.attach_idx(idx)
        return mon

    def _monomial_from_atoms(self,
                             atoms: List[InternalAtomicMonomial]
                             ) -> CompoundMoment:
        """Build an instance of `CompoundMonomial` from a list of instances
        of `InternalAtomicMonomial`.

        Parameters
        ----------
        atoms : List[InternalAtomicMonomial]
            List of instances of `InternalAtomicMonomial`.

        Returns
        -------
        CompoundMoment
            A `CompoundMonomial` with atomic factors given by `atoms`.
        """
        key = tuple(sorted(atoms))
        try:
            return self.monomial_from_atoms[key]
        except KeyError:
            mon = CompoundMoment(atoms)
            try:
                mon.idx = self.first_free_idx
                self.first_free_idx += 1
            except AttributeError:
                pass
            self.monomial_from_atoms[key] = mon
            self.monomial_from_name[mon.name] = mon
            self.monomial_from_name[mon.legacy_name] = mon  # For legacy compatibility!
            self.monomial_from_symbol[mon.symbol] = mon
            return mon

    def _sanitise_monomial(self, mon: Any) -> CompoundMoment:
        """Return a ``CompoundMonomial`` built from ``mon``, where ``mon`` can
        be either the name of a moment as a string, a SymPy variable, a
        monomial encoded as a 2D array, or an integer in case the moment is the
        unit moment or the zero moment.


        Parameters
        ----------
        mon : Any
            The name of a moment as a string, a SymPy variable with the name of
            a valid moment, a 2D array encoding of a moment or an integer in
            case the moment is the unit moment or the zero moment.

        Returns
        -------
        CompoundMoment
            Instance of ``CompoundMonomial`` built from ``mon``.

        Raises
        ------
        Exception
            If ``mon`` is the constant monomial, it can only be numbers 0 or 1
        Exception
            If the type of ``mon`` is not supported.
        """
        if isinstance(mon, CompoundMoment):
            return mon
        elif isinstance(mon, InternalAtomicMonomial):
            return self._monomial_from_atoms([mon])
        elif isinstance(mon, (sp.core.symbol.Symbol,
                              sp.core.power.Pow,
                              sp.core.mul.Mul,
                              sp.core.symbol.Expr)):
            try:
                return self.monomial_from_symbol[mon]
            except KeyError:
                symbols = flatten_symbolic_powers(mon)
                if len(symbols) == 1:
                    try:
                        return self.monomial_from_name[str(symbols[0])]
                    except KeyError:
                        pass
                array = np.concatenate([self._interpret_atomic_string(str(op))
                                        for op in symbols])
                return self._sanitise_monomial(array)
        elif isinstance(mon, (tuple, list, np.ndarray)):
            array = np.asarray(mon, dtype=self.np_dtype)
            assert array.ndim <= 2, \
                "The monomial representations must be 1d or 2d arrays."
            if array.ndim == 2:
                return self._sanitise_monomial(self.mon_to_lexrepr(array))
            elif array.ndim == 1:
                return self.Monomial(array)
            else:
                assert array.ndim == 2, \
                    "The monomial representations must be 2d arrays."
        elif isinstance(mon, str):
            try:
                return self.monomial_from_name[mon]
            except KeyError:
                eprint(f"As of now we only recognize \n{list(self.monomial_from_name.keys())}")
                return self._sanitise_monomial(self._interpret_name(mon))
        elif isinstance(mon, Real):
            if np.isclose(float(mon), 1):
                return self.One
            else:
                raise Exception(f"Constant monomial {mon} can only be 1.")
        else:
            raise Exception(f"sanitise_monomial: {mon} is of type " +
                            f"{type(mon)} and is not supported.")

    def _sanitise_dict(self, input_dict: Any) -> Dict:
        """Sanitize a dictionary of monomials.

        Parameters
        ----------
        input_dict : Any
            The dictionary to be sanitized.

        Returns
        -------
        Dict
            The sanitized dictionary.
        """
        if isinstance(input_dict, sp.core.expr.Expr):
            if input_dict.free_symbols:
                input_dict_copy = {k: float(v) for k, v in sp.expand(input_dict).as_coefficients_dict().items()}
            else:
                input_dict_copy = {}
        else:
            input_dict_copy = input_dict
        output_dict = defaultdict(int)
        for k, v in input_dict_copy.items():
            if not np.isclose(v, 0):
                output_dict[self._sanitise_monomial(k)] += v
        return output_dict

    ###########################################################################
    # ROUTINES RELATED TO NAME PARSING                                        #
    ###########################################################################
    def _interpret_name(self,
                        monomial: Union[str, sp.core.symbol.Expr, int]
                        ) -> np.ndarray:
        """Build a 2D array encoding of a monomial which can be passed either
        as a string, as a SymPy expression or as an integer.

        Parameters
        ----------
        monomial : Union[str, sympy.core.symbol.Expr, int]
            Input moment.

        Returns
        -------
        numpy.ndarray
            2D array encoding of the input moment.
        """
        if str(monomial) == '1':
            return self.identity_operator
        elif isinstance(monomial, str):
            try:
                return self.monomial_from_name[monomial]
            except KeyError:
                factors = monomial.split("*")
        elif isinstance(monomial, tuple) or isinstance(monomial, list):
            factors = [str(factor) for factor in monomial]
        elif isinstance(monomial, sp.core.symbol.Expr):
            try:
                return self.monomial_from_symbol[monomial]
            except KeyError:
                factors = [str(factor)
                           for factor in flatten_symbolic_powers(monomial)]
        else:
            raise Exception(f'Cannot interpret monomial with name {monomial} of type {type(monomial)}')
        return np.vstack(tuple(self._interpret_atomic_string(factor_string)
                               for factor_string in factors))

    def _interpret_atomic_string(self, factor_string: str) -> np.ndarray:
        """Build a 2D array encoding of a moment that cannot be further
        factorised into products of other moments.

        Parameters
        ----------
        factor_string : str
            String representation of a moment in expected value notation, e.g.,
            ``"<A_1_1_1_2*B_2_1_3_4>"``.

        Returns
        -------
        numpy.ndarray
            2D array encoding of the input atomic moment.
        """
        try:
            return self.atom_from_name[factor_string].as_2d_array
        except (KeyError, AttributeError):
            pass
        assert ((factor_string[0] == "<" and factor_string[-1] == ">")
                or (factor_string[0:1] == "P[" and factor_string[-1] == "]")
                or set(factor_string).isdisjoint(set("| "))), \
            ("Monomial names must be between < > signs, or in conditional " +
             f"probability form, whereas input received was {factor_string}")
        if factor_string[-1] in {'>', ')', "]", "}"}:
            cleaned_factor_string = factor_string[:-1]
            substrings_to_kill = {"P[", "P(", "p[", "p(", "<"}
            for substring in substrings_to_kill:
                cleaned_factor_string = cleaned_factor_string.replace(substring, '')
            cleaned_factor_string = cleaned_factor_string.replace( ' & ',' ')
            operators = cleaned_factor_string.split(" ")
            return np.vstack(tuple(self._interpret_operator_string(op_string)
                                   for op_string in operators))
        else:
            return self._interpret_operator_string(factor_string)[np.newaxis]

    def _interpret_operator_string(self, op_string: str) -> np.ndarray:
        """Build a 1D array encoding of an operator passed as a string.

        Parameters
        ----------
        op_string : str
            String representation of an operator, e.g., ``"B_2_1_3_4"``.

        Returns
        -------
        numpy.ndarray
            2D array encoding of the operator.
        """
        return self._lexorder[self.op_from_name[op_string]]

    ###########################################################################
    # ROUTINES RELATED TO THE GENERATION OF THE LP                            #
    ###########################################################################
    def _generate_lp(self) -> None:
        """Creates the LP associated with the inflation problem.

        In the inflated graph there are many symmetries coming from invariance
        under swaps of the copied sources, which are used to remove variables
        from the LP.
        """
        # Note that there IS NO POSSIBILITY OF ORTHOGONALITY in the LP
        # and hence we start indices from zero, always.
        try:
            self.generate_lp_has_been_called += 1
        except AttributeError:
            self.generate_lp_has_been_called = 0
        if self.generate_lp_has_been_called:
            return None

        self._atomic_monomial_from_hash = {}
        self.monomial_from_atoms = {}
        self.Constant_Term = self.One.__copy__()
        self.Constant_Term.name = self.constant_term_name
        self.monomial_from_name[self.constant_term_name] = self.Constant_Term

        self.raw_n_columns = len(self._raw_monomials_as_lexboolvecs)
        self.raw_n_columns_non_CG = len(self._raw_monomials_as_lexboolvecs_non_CG)

        self._raw_lookup_dict = {bitvec.tobytes(): i for i, bitvec in
                                 enumerate(self._raw_monomials_as_lexboolvecs)}

        symmetrization_required = np.any(self.inflation_levels - 1)
        if symmetrization_required:
            # Calculate the inflation symmetries
            if self.verbose > 0:
                eprint("Initiating symmetry calculation...")
            orbits = self._discover_inflation_orbits(self._raw_monomials_as_lexboolvecs,
                                                     raw_hash_table=self._raw_lookup_dict)
            if self.verbose > 1:
                eprint("Halfway through symmetry calculations...")
            old_reps_CG, unique_indices_CG, inverse_CG = np.unique(
                orbits,
                return_index=True,
                return_inverse=True)
            self.num_CG = len(old_reps_CG)

            orbits_non_CG = self._discover_inflation_orbits(
                self._raw_monomials_as_lexboolvecs_non_CG)
            old_reps_non_CG, unique_indices_non_CG, inverse_non_CG = np.unique(
                orbits_non_CG, return_index=True, return_inverse=True)
            self.num_non_CG = len(old_reps_non_CG)
            if self.verbose > 1:
                eprint(f"Orbits discovered! {self.num_CG} unique monomials.")
            # Obtain the real generating monomials after accounting for symmetry
        else:
            self.num_CG = self.raw_n_columns
            unique_indices_CG = np.arange(self.num_CG)
            inverse_CG = unique_indices_CG
            self.num_non_CG = self.raw_n_columns_non_CG
            unique_indices_non_CG = np.arange(self.num_non_CG)
        self.inverse = inverse_CG

        self._monomials_as_lexboolvecs = self._raw_monomials_as_lexboolvecs[unique_indices_CG]
        self._monomials_as_lexboolvecs_non_CG = self._raw_monomials_as_lexboolvecs_non_CG[unique_indices_non_CG]
        self.n_columns = len(self._monomials_as_lexboolvecs)

        self.nof_collins_gisin_inequalities = self.num_non_CG

        if self.verbose > 0:
            eprint("Number of variables in the LP:",
                  self.n_columns)
            eprint("Number of nontrivial inequality constraints in the LP:",
                    self.nof_collins_gisin_inequalities)

        # Associate Monomials to the remaining entries.
        _compmonomial_to_idx = dict()
        self.extra_inverse = np.arange(self.n_columns, dtype=int)
        first_free_index = 0
        for idx, mon_as_lexboolvec in tqdm(enumerate(self._monomials_as_lexboolvecs),
                             disable=not self.verbose,
                             desc="Initializing monomials   ",
                             total=self.n_columns):
            mon = self.Monomial(np.flatnonzero(mon_as_lexboolvec), first_free_index)
            try:
                current_index = _compmonomial_to_idx[mon]
                mon.idx = current_index
            except KeyError:
                current_index = first_free_index
                _compmonomial_to_idx[mon] = current_index
                first_free_index += 1
            self.extra_inverse[idx] = current_index
        self.inverse = self.extra_inverse[self.inverse] # Hack to allow for powerful symmetries

        monomials_as_list = list(_compmonomial_to_idx.keys())
        self.monomials = np.array(monomials_as_list, dtype = object)
        # assert np.array_equal(list(_compmonomial_to_idx.values()), np.arange(len(self.monomials))), "Something went wrong with monomial initialization."
        old_num_columns = self.n_columns
        self.n_columns = len(self.monomials)
        self.first_free_idx = first_free_index
        self.monomial_names = np.array([mon.name for mon in monomials_as_list], dtype=object)
        if self.n_columns < old_num_columns:
            if self.verbose > 0:
                eprint("Further variable reduction has been made possible. Number of variables in the LP:",
                       self.n_columns)
        self.compmonomial_from_idx = dict(zip(range(self.n_columns), monomials_as_list))
        self.compmonomial_to_idx = dict(zip(monomials_as_list, range(self.n_columns)))
        del _compmonomial_to_idx, monomials_as_list
        collect(generation=2)
        assert self.monomials[0] == self.One, f"Sparse indexing requires that first column represent one, but got {self.monomials[0]}"

        # assert len(self.compmonomial_to_idx.keys()) == self.n_columns, \
        #     (f"Multiple indices are being associated to the same monomial. \n" +
        #     f"Expected {self.n_columns}, got {len(self.compmonomial_to_idx.keys())}.")


        _counter = Counter([mon.knowability_status
                            for mon in self.monomials])
        self.n_knowable           = _counter["Knowable"]
        self.n_something_knowable = _counter["Semi"]
        self.n_unknowable         = _counter["Unknowable"]
        if self.verbose > 1:
            eprint(f"The problem has {self.n_knowable} knowable monomials, " +
                  f"{self.n_something_knowable} semi-knowable monomials, " +
                  f"and {self.n_unknowable} unknowable monomials.")

        # This dictionary useful for certificates_as_probs
        self.names_to_symbols = {mon.name: mon.symbol
                                 for mon in self.monomials}
        self.names_to_symbols[self.constant_term_name] = sp.S.One

        self._set_lowerbounds(None)
        self._set_upperbounds(None)
        self.set_objective(None)
        self.set_values(None)

        self._lp_has_been_generated = True
        if self.verbose > 1:
            eprint("LP initialization complete, ready to accept further specifics.")

    def _template_to_event_boolarray(self, template: List[int], decompressor: List[np.ndarray]) -> np.ndarray:
        if template:
            to_expand = partsextractor(decompressor, sorted(template, key=op_neg))
            return reduce(nb_outer_bitwise_or, to_expand)
        else:
            return self.blank_bool_array

    @cached_property
    def all_and_maximal_compatible_templates(self):
        return self.InflationProblem.all_and_maximal_compatible_templates()
    @property
    def all_compatible_templates(self):
        return self.all_and_maximal_compatible_templates[0]
    @property
    def maximal_compatible_templates(self):
        return self.all_and_maximal_compatible_templates[1]

    @cached_property
    def _raw_monomials_as_lexboolvecs(self) -> np.ndarray:
        r"""Creates the generating set of monomials (as boolvecs),
        of ALL lengths, limited to Collins-Gisin notation.
        """
        return np.vstack([
            self._template_to_event_boolarray(subclique, self._CG_limited_ortho_groups_as_boolarrays)
            for subclique in self.all_compatible_templates
        ])

    @cached_property
    def _raw_monomials_as_lexboolvecs_non_CG(self) -> np.ndarray:
        r"""Creates the generating set of monomials (as boolvecs),
        of ALL lengths, limited to Collins-Gisin notation.
        """
        return np.vstack([
            self._template_to_event_boolarray(clique, self._all_ortho_groups_as_boolarrays)
            for clique in self.maximal_compatible_templates
        ])

    @cached_property
    def minimal_sparse_equalities(self) -> coo_array:
        """Given the generating monomials, infer conversion to Collins-Gisin 
        notation."""
        eq_row, eq_col, eq_data = [], [], []
        nof_equalities = 0
        if np.any(self._boolvec_for_FR_eqs):
            alternatives_as_boolarrays = {v: np.pad(r[1:], ((1, 0), (0, 0)))
                                          for v, r in zip(
                    np.flatnonzero(self._boolvec_for_FR_eqs).flat,
                    self._CG_nonadjusting_ortho_groups_as_boolarrays)}
            alternatives_as_signs = {
                i: np.count_nonzero(bool_array, axis=1).astype(bool)
                for i, bool_array in alternatives_as_boolarrays.items()}

            for bool_vec in tqdm(self._monomials_as_lexboolvecs,
                    disable=not self.verbose,
                    desc="Discovering equalities   "):
                critical_boolvec_intersection = np.bitwise_and(bool_vec, self._boolvec_for_FR_eqs)
                if np.any(critical_boolvec_intersection):

                    absent_c_boolvec = bool_vec.copy()
                    absent_c_boolvec[critical_boolvec_intersection] = False
                    critical_values_in_boovec = np.flatnonzero(critical_boolvec_intersection)
                    signs = reduce(nb_outer_bitwise_xor,
                                   (alternatives_as_signs[i] for i in critical_values_in_boovec.flat))
                    adjustments = reduce(nb_outer_bitwise_or,
                                   (alternatives_as_boolarrays[i] for i in critical_values_in_boovec.flat))
                    terms_as_boolvecs = np.bitwise_or(
                        absent_c_boolvec[np.newaxis],
                        adjustments)
                    # Conversion from inequality to equality:
                    signs = np.hstack((signs,1))
                    terms_as_boolvecs = np.vstack((terms_as_boolvecs, 
                                                   bool_vec))
                    terms_as_rawidx = [self._raw_lookup_dict[boolvec.tobytes()] 
                                       for boolvec in terms_as_boolvecs]
                    terms_as_idxs = self.inverse[terms_as_rawidx]
                    true_signs = np.power(-1, signs)

                    eq_row.extend([nof_equalities] * len(signs))
                    eq_col.extend(terms_as_idxs.flat)
                    eq_data.extend(true_signs.flat)
                    nof_equalities += 1
            if self.verbose > 0:
                eprint("Number of nontrivial equality constraints in the LP:",
                        nof_equalities)
        return coo_array((eq_data, (eq_row, eq_col)),
                          shape=(nof_equalities, self.n_columns))

    @property
    def sparse_extra_equalities(self) -> coo_array:
        """Extra equalities in sparse matrix form."""
        eq_row, eq_col, eq_data = [], [], []
        nof_equalities = len(self.extra_equalities)
        for row_idx, eq in enumerate(self.extra_equalities):
            nof_vars = len(eq)
            eq_row.extend(np.repeat(row_idx, nof_vars))
            eq_col.extend([self.compmonomial_to_idx[x] for x in eq])
            eq_data.extend(eq.values())
        return coo_array((eq_data, (eq_row, eq_col)),
                          shape=(nof_equalities, self.n_columns))

    @property
    def sparse_equalities(self) -> coo_array:
        """All equalities (minimal and extra) in sparse matrix 
        form."""
        return vstack((self.minimal_sparse_equalities,
                       self.sparse_extra_equalities))

    @cached_property
    def minimal_sparse_inequalities(self) -> coo_array:
        """Here we express the nonnegativity of all `global` (maximal number
         of variables) events, converting the expressions into Collins-Gisin
         notation as needed."""
        ineq_row, ineq_col, ineq_data = [], [], []
        nof_inequalities = 0
        alternatives_as_boolarrays = {v: np.pad(r[:-1], ((1, 0), (0, 0)))
                                      for v, r in zip(
                np.flatnonzero(self._boolvec_for_CG_ineqs).flat,
                self._CG_adjusting_ortho_groups_as_boolarrays)}
        alternatives_as_signs = {
            i: np.count_nonzero(bool_array, axis=1).astype(bool)
            for i, bool_array in alternatives_as_boolarrays.items()}
        for bool_vec in tqdm(self._monomials_as_lexboolvecs_non_CG,
                             disable=not self.verbose,
                             desc="Discovering inequalities   "):
            critical_boolvec_intersection = np.bitwise_and(bool_vec,
                                                           self._boolvec_for_CG_ineqs)
            if critical_boolvec_intersection.any():
                absent_c_boolvec = bool_vec.copy()
                absent_c_boolvec[critical_boolvec_intersection] = False
                critical_values_in_boovec = np.flatnonzero(
                    critical_boolvec_intersection)
                signs = reduce(nb_outer_bitwise_xor,
                               (alternatives_as_signs[i] for i in
                                critical_values_in_boovec.flat))
                adjustments = reduce(nb_outer_bitwise_or,
                                     (alternatives_as_boolarrays[i] for i in
                                      critical_values_in_boovec.flat))
                terms_as_boolvecs = np.bitwise_or(
                    absent_c_boolvec[np.newaxis],
                    adjustments)
                terms_as_rawidx = [self._raw_lookup_dict[term_boolvec.tobytes()]
                                   for term_boolvec in terms_as_boolvecs]
                terms_as_idxs = self.inverse[terms_as_rawidx]
                true_signs = np.power(-1, signs)

                ineq_row.extend([nof_inequalities] * len(signs))
                ineq_col.extend(terms_as_idxs.flat)
                ineq_data.extend(true_signs.flat)
            else:
                ineq_row.append(nof_inequalities)
                ineq_col.append(self.inverse[self._raw_lookup_dict[bool_vec.tobytes()]])
                ineq_data.append(1)
            nof_inequalities += 1
        return coo_array((ineq_data, (ineq_row, ineq_col)),
                          shape=(nof_inequalities, self.n_columns))

    @property
    def sparse_extra_inequalities(self) -> coo_array:
        """Extra inequalities in sparse matrix form."""
        ineq_row, ineq_col, ineq_data = [], [], []
        nof_inequalities = len(self.extra_inequalities)
        for row_idx, ineq in enumerate(self.extra_inequalities):
            nof_vars = len(ineq)
            ineq_row.extend(np.repeat(row_idx, nof_vars))
            ineq_col.extend([self.compmonomial_to_idx[x] for x in ineq])
            ineq_data.extend(ineq.values())
        return coo_array((ineq_data, (ineq_row, ineq_col)),
                          shape=(nof_inequalities, self.n_columns))

    @property
    def sparse_inequalities(self) -> coo_array:
        """All inequalities (minimal and extra) in sparse matrix form."""
        return vstack((self.minimal_sparse_inequalities,
                       self.sparse_extra_inequalities))

    def _coo_vec_to_mon_dict(self,
                             col: np.ndarray,
                             data: np.ndarray) -> Dict:
        """Convert a COO vector to a dictionary of monomials.
        
        Parameters
        ----------
        col : numpy.ndarray
            Column indices of the COO vector.
        data : numpy.ndarray
            Data of the COO vector.
            
        Returns
        -------
        Dict
            Dictionary of monomials.
        """
        return dict(zip(self.monomials[col].flat, data))

    def _coo_vec_to_name_dict(self, 
                              col: np.ndarray,
                              data: np.ndarray) -> Dict:
        """Convert a COO vector to a dictionary of monomial names.
        
        Parameters
        ----------
        col : numpy.ndarray
            Column indices of the COO vector.
        data : numpy.ndarray
            Data of the COO vector.
        
        Returns
        -------
        Dict
            Dictionary of monomial names.
        """
        return dict(zip(self.monomial_names[col].flat, data))

    def _coo_mat_to_dict(self,
                         input_coo_mat: coo_array,
                         string_keys: bool = False) -> List[Dict]:
        """Convert a COO matrix to a list of dictionaries.
        
        Parameters
        ----------
        input_coo_mat : coo_array
            Input COO matrix.
        string_keys : bool, optional
            Whether to use string keys or not, by default, ``False``.
        
        Returns
        -------
        List[Dict]
            List of dictionaries.
        """
        input_lil_mat = input_coo_mat.tolil(copy=False)
        args_iter = zip(input_lil_mat.rows, input_lil_mat.data)
        if string_keys:
            return [self._coo_vec_to_name_dict(*args) for args in args_iter]
        else:
            return [self._coo_vec_to_mon_dict(*args) for args in args_iter]

    def _mon_dict_to_coo_vec(self, monomials_dict: Dict) -> coo_array:
        """
        This is a PLACEHOLDER function, possibly to be deprecated, to convert
         dicts into COO matrices.
        """
        data = list(monomials_dict.values())
        keys = list(monomials_dict.keys())
        col = partsextractor(self.compmonomial_to_idx, keys)
        row = np.zeros(len(col), dtype=int)
        return coo_array((data, (row, col)), shape=(1, self.n_columns))

    @property
    def moment_equalities(self):
        """List of dictionaries of moment equalities."""
        return self._coo_mat_to_dict(self.sparse_equalities)

    @property
    def moment_inequalities(self):
        """List of dictionaries of moment inequalities."""
        return self._coo_mat_to_dict(self.sparse_inequalities)


    def _discover_inflation_orbits(self, 
                                   _raw_monomials_as_lexboolvecs: np.ndarray,
                                   raw_hash_table=False
                                   ) -> np.ndarray:
        """Calculates all the symmetries pertaining to the set of generating
        monomials. The new set of operators is a permutation of the old. The
        function outputs a list of all permutations.

        Returns
        -------
        numpy.ndarray[int]
            The orbits of the generating columns implied by the inflation
            symmetries.
        """
        nof_bitvecs_to_parse = len(_raw_monomials_as_lexboolvecs)
        if len(self.lexorder_symmetries) > 1:
            if not raw_hash_table:
                hash_table = {bitvec.tobytes(): i for i, bitvec in
                              enumerate(_raw_monomials_as_lexboolvecs)}
            else:
                hash_table = raw_hash_table
            non_identity_symmetries = self.lexorder_symmetries[1:]
            orbits = np.full(nof_bitvecs_to_parse, -1, dtype=int)
            for i in tqdm(range(nof_bitvecs_to_parse),
                          disable=not self.verbose,
                          total=nof_bitvecs_to_parse,
                          desc="Calculating orbits...             "):
                if orbits[i] == -1:
                    orbits[i] = i
                    initial = _raw_monomials_as_lexboolvecs[i]
                    initial_hash = initial.tobytes()
                    variants = initial[non_identity_symmetries]
                    variant_hashes = {variant.tobytes() for variant in variants}.difference({initial_hash})
                    remaining_orbit = []
                    for variant_hash in variant_hashes:
                        try:
                            remaining_orbit.append(hash_table[variant_hash])
                        except KeyError:
                            continue
                    orbits[remaining_orbit] = i
            return orbits
        else:
            return np.arange(nof_bitvecs_to_parse, dtype=int)

    ###########################################################################
    # HELPER FUNCTIONS FOR ENSURING CONSISTENCY                               #
    ###########################################################################
    def _cleanup_after_set_values(self) -> None:
        """Helper function to reset or make consistent class attributes after
        setting values."""
        if self.supports_problem:
            # Add lower bounds to monomials inside the support
            nonzero_known_monomials = [mon for
                                       mon, value in self.known_moments.items()
                                       if not np.isclose(value, 0)]
            for mon in nonzero_known_monomials:
                self.moment_lowerbounds[mon] = 1.
                del self.known_moments[mon]
            self.semiknown_moments = {}
        num_nontrivial_known = len(self.known_moments)
        if self.verbose > 1 and num_nontrivial_known > 1:
            eprint("Number of variables with fixed numeric value:",
                  num_nontrivial_known)
        if len(self.semiknown_moments):
            for k in self.known_moments.keys():
                self.semiknown_moments.pop(k, None)
        num_semiknown = len(self.semiknown_moments)
        if self.verbose > 1 and num_semiknown > 0:
            eprint(f"Number of semiknown variables: {num_semiknown}")

    def _reset_bounds(self) -> None:
        """Reset the lists of bounds."""
        self._reset_lowerbounds()
        self._reset_upperbounds()
        collect()

    def _reset_lowerbounds(self) -> None:
        """Reset the list of lower bounds."""
        self._reset_solution()
        self.moment_lowerbounds = {}

    def _reset_upperbounds(self) -> None:
        """Reset the list of upper bounds."""
        self._reset_solution()
        self.moment_upperbounds = {}

    def _reset_objective(self) -> None:
        """Reset the objective function."""
        self._reset_solution()
        self.objective = defaultdict(int)
        self.maximize = True  # Direction of the optimization

    def _reset_values(self) -> None:
        """Reset the known values."""
        self._reset_solution()
        self.known_moments     = {}
        self.semiknown_moments = {}
        self.known_moments[self.One] = 1.
        self.extra_equalities = []
        self.extra_inequalities = []
        collect()

    def _reset_solution(self) -> None:
        """Resets class attributes storing the solution to the LP
        relaxation."""
        for attribute in {"primal_objective",
                          "objective_value",
                          "solution_object"}:
            try:
                delattr(self, attribute)
            except AttributeError:
                pass
        self.status = "Not yet solved"


    ###########################################################################
    # Preparation for passing to external interfaces                          #
    ###########################################################################

    @cached_property
    def moment_equalities_by_name(self):
        """List of dictionaries of moment equalities by name."""
        return self._coo_mat_to_dict(self.sparse_equalities, string_keys=True)

    @cached_property
    def moment_inequalities_by_name(self):
        """List of dictionaries of moment inequalities by name."""
        return self._coo_mat_to_dict(self.sparse_inequalities, 
                                     string_keys=True)

    @cached_property
    def factorization_conditions_by_name(self):
        """Dictionary of factorization conditions by name."""
        return {mon.name: tuple(fac.name for fac in val)
                for mon, val in self.factorization_conditions.items()}

    @cached_property
    def sparse_factorization_conditions(self):
        """Dictionary of factorization conditions in sparse matrix form."""
        return {self.compmonomial_to_idx[mon]: partsextractor(self.compmonomial_to_idx, val)
                for mon, val in self.factorization_conditions.items()}

    @cached_property
    def quadratic_factorization_conditions_by_name(self):
        """Dictionary of quadratic factorization conditions by name."""
        return {mon.name: tuple(fac.name for fac in val) 
                for mon, val in self.quadratic_factorization_conditions.items()}

    @cached_property
    def sparse_quadratic_factorization_conditions(self):
        """Dictionary of quadratic factorization conditions in sparse matrix
        form."""
        return {self.compmonomial_to_idx[mon]: partsextractor(self.compmonomial_to_idx, val) 
                for mon, val in self.quadratic_factorization_conditions.items()}

    @property
    def sparse_objective(self) -> coo_array:
        """Sparse matrix representation of the objective function."""
        # TO BE DEPRECATED
        return self._mon_dict_to_coo_vec(self.objective)
    
    @property
    def objective_by_name(self) -> Dict:
        """Dictionary representation of the objective function."""
        return self._coo_mat_to_dict(self.sparse_objective,
                                     string_keys=True)[0]
    
    @property
    def sparse_known_vars(self) -> coo_array:
        """Sparse matrix representation of the known values."""
        # TO BE DEPRECATED
        return self._mon_dict_to_coo_vec(self.known_moments)
    
    @property
    def known_vars_by_name(self) -> Dict:
        """Dictionary representation of the known values."""
        return self._coo_mat_to_dict(self.sparse_known_vars, 
                                     string_keys=True)[0]
        
    @property
    def sparse_lowerbounds(self) -> coo_array:
        """Sparse matrix representation of the lower bounds."""
        # TO BE DEPRECATED
        return self._mon_dict_to_coo_vec(self.moment_lowerbounds)
    
    @property
    def lowerbounds_by_name(self) -> Dict:
        """Dictionary representation of the lower bounds."""
        return self._coo_mat_to_dict(self.sparse_lowerbounds, 
                                     string_keys=True)[0]
        
    @property
    def sparse_upperbounds(self) -> coo_array:
        """Sparse matrix representation of the upper bounds."""
        # TO BE DEPRECATED
        return self._mon_dict_to_coo_vec(self.moment_upperbounds)
    
    @property
    def upperbounds_by_name(self) -> Dict:
        """Dictionary representation of the upper bounds."""
        return self._coo_mat_to_dict(self.sparse_upperbounds, 
                                     string_keys=True)[0]

    @property
    def sparse_semiknown(self) -> coo_array:
        """Sparse matrix representation of the semiknown values."""
        nof_semiknown = len(self.semiknown_moments)
        nof_variables = len(self.compmonomial_to_idx)
        row = np.repeat(np.arange(nof_semiknown), 2)
        col = [(self.compmonomial_to_idx[x], self.compmonomial_to_idx[x2])
               for x, (c, x2) in self.semiknown_moments.items()]
        col = list(sum(col, ()))
        data = [(1, -c) for x, (c, x2) in self.semiknown_moments.items()]
        data = list(sum(data, ()))
        return coo_array((data, (row, col)),
                          shape=(nof_semiknown, nof_variables))

    @property
    def semiknown_by_name(self) -> Dict:
        """Dictionary representation of the semiknown values."""
        return {mon.name: (coeff, subs.name)
                                         for mon, (coeff, subs)
                                         in self.semiknown_moments.items()}

    def _prepare_solver_matrices(self,
                                 separate_bounds: bool = True) -> dict:
        """Convert arguments from dictionaries to sparse coo_array form to
        pass to the solver.

        Parameters
        ----------
        separate_bounds : bool, optional
            Whether to have variable bounds as a separate item in
            ``solverargs`` (True) or absorb them into the inequalities (False).
            By default, ``True``.

        Returns
        -------
        dict
            The arguments to be passed to the solver.

        Raises
        ------
        Exception
            If the LP has not been generated yet.
        """
        if not self._lp_has_been_generated:
            raise Exception("LP is not generated yet. " +
                            "Call \"InflationLP._generate_lp()\" first")

        assert set(self.known_moments.keys()).issubset(self.monomials), \
            ("Error: Tried to assign known values outside of the variables: " +
             str(set(self.known_moments.keys()
                     ).difference(self.monomials)))

        internal_equalities = vstack((self.sparse_equalities,
                                      self.sparse_semiknown))

        solverargs = {"objective": self.sparse_objective,
                      "known_vars": self.sparse_known_vars,
                      "equalities": internal_equalities,
                      "inequalities": self.sparse_inequalities,
                      "variables": self.monomial_names}
        if separate_bounds:
            solverargs["lower_bounds"] = self.sparse_lowerbounds
            solverargs["upper_bounds"] = self.sparse_upperbounds
        else:
            lb_mat = expand_sparse_vec(self.sparse_lowerbounds,
                                       conversion_style="lb")
            ub_mat = expand_sparse_vec(self.sparse_upperbounds,
                                       conversion_style="ub")
            solverargs["inequalities"] = vstack(
                (self.sparse_inequalities, lb_mat, ub_mat))
        return solverargs

    def _prepare_solver_arguments(self, 
                                  separate_bounds: bool = True) -> dict:
        """Prepare arguments to pass to the solver.

        The solver takes as input the following arguments, which are all
        dicts with keys as scalar LP variables:
            * "objective": dict with values the coefficient of the key
            variable in the objective function.
            * "known_vars": scalar variables that are fixed to be constant.
            * "semiknown_vars": if applicable, linear proportionality
            constraints between variables in the LP.
            * "equalities": list of dicts where each dict gives the
            coefficients of the keys in a linear equality constraint.
            * "inequalities": list of dicts where each dict gives the
            coefficients of the keys in a linear inequality constraint.

        Parameters
        ----------
        separate_bounds : bool, optional
            Whether to have variable bounds as a separate item in
            ``solverargs`` (True) or absorb them into the inequalities (False).
            By default, ``True``.

        Returns
        -------
        dict
            A tuple with the arguments to be passed to the solver.

        Raises
        ------
        Exception
            If the LP relaxation has not been calculated yet.
        """
        if not self._lp_has_been_generated:
            raise Exception("LP is not generated yet. " +
                            "Call \"InflationLP._generate_lp()\" first")

        assert set(self.known_moments.keys()).issubset(self.monomials), \
            ("Error: Tried to assign known values outside of the variables: " +
             str(set(self.known_moments.keys()
                     ).difference(self.monomials)))

        # Rectification in the event of unnormalized problem
        variables = self.monomial_names.tolist()
        if {1, self.constant_term_name}.isdisjoint(self.known_vars_by_name):
            self.known_vars_by_name[self.constant_term_name] = 1.
            if self.constant_term_name not in variables:
                variables.append(self.constant_term_name)
        solverargs = {"objective": self.objective_by_name,
                      "known_vars": self.known_vars_by_name,
                      "semiknown_vars": self.semiknown_by_name,
                      "equalities": self.moment_equalities_by_name,
                      "inequalities": self.moment_inequalities_by_name,
                      "variables": variables}
        if separate_bounds:
            solverargs["lower_bounds"] = self.lowerbounds_by_name
            solverargs["upper_bounds"] = self.upperbounds_by_name
        else:
            solverargs["inequalities"].extend({mon.name: 1, '1': -bnd}
                                              for mon, bnd in
                                              self.moment_lowerbounds.items())
            solverargs["inequalities"].extend({mon.name: -1, '1': bnd}
                                              for mon, bnd in
                                              self.moment_upperbounds.items())
        return solverargs

    ###########################################################################
    # OTHER ROUTINES                                                          #
    ###########################################################################
    def _atomic_knowable_q(self, atomic_monarray: np.ndarray) -> bool:
        """Return ``True`` if the input monomial, encoded as a 2D array,
        can be associated to a knowable value in the scenario, and ``False``
        otherwise.

        Parameters
        ----------
        atomic_monarray : numpy.ndarray
            Monomial encoded as a 2D array.

        Returns
        -------
        bool
            Whether the monomial could be assigned a numerical value.
        """
        if not is_knowable(atomic_monarray):
            return False
        elif self.network_scenario:
            return True
        else:
            return self._is_knowable_q_non_networks(np.take(atomic_monarray,
                                                            [0, -2, -1],
                                                            axis=1))

    def _set_upperbounds(self, upperbounds: Union[dict, None]) -> None:
        """Set upper bounds for variables in the LP relaxation.

        Parameters
        ----------
        upperbounds : Union[dict, None]
            Dictionary with keys as moments and values as upper bounds. The
            keys can be either strings, instances of `CompoundMonomial` or
            moments encoded as 2D arrays.
        """
        self._reset_upperbounds()
        if upperbounds is None:
            return
        sanitized_upperbounds = {}
        for mon, upperbound in upperbounds.items():
            mon = self._sanitise_monomial(mon)
            if mon not in sanitized_upperbounds.keys():
                sanitized_upperbounds[mon] = upperbound
            else:
                old_bound = sanitized_upperbounds[mon]
                assert np.isclose(old_bound,
                                  upperbound), \
                    (f"Contradiction: Cannot set the same monomial {mon} to " +
                     "have different upper bounds.")
        self.moment_upperbounds = sanitized_upperbounds

    def _set_lowerbounds(self, lowerbounds: Union[dict, None]) -> None:
        """Set lower bounds for variables in the LP relaxation.

        Parameters
        ----------
        lowerbounds : Union[dict, None]
            Dictionary with keys as moments and values as upper bounds. The
            keys can be either strings, instances of `CompoundMonomial` or
            moments encoded as 2D arrays.
        """
        self._reset_lowerbounds()
        if lowerbounds is None:
            return
        sanitized_lowerbounds = {}
        for mon, lowerbound in lowerbounds.items():
            mon = self._sanitise_monomial(mon)
            if mon not in sanitized_lowerbounds.keys():
                sanitized_lowerbounds[mon] = lowerbound
            else:
                old_bound = sanitized_lowerbounds[mon]
                assert np.isclose(old_bound, lowerbound), \
                    (f"Contradiction: Cannot set the same monomial {mon} to " +
                     "have different lower bounds.")
        self.moment_lowerbounds = sanitized_lowerbounds

    def mon_to_boolvec(self, mon: np.ndarray) -> np.ndarray:
        """Convert a monomial to a boolean vector.
        
        Parameters
        ----------
        mon : numpy.ndarray
            Monomial as a 2D array.
        
        Returns
        -------
        numpy.ndarray
            Boolean vector.
        """
        boolvec = self.blank_bool_vec.copy()
        boolvec[self.mon_to_lexrepr(mon)] = True
        return boolvec


"""
This file contains auxiliary functions of general purpose

@authors: Emanuel-Cristian Boghiu, Elie Wolfe and Alejandro Pozas-Kerstjens
"""
from __future__ import print_function
import sympy

import numpy as np
import scipy.sparse as sps

from itertools import chain
from typing import Any, Dict, Iterable, List, Tuple, Union
from sys import stderr
from operator import itemgetter
from collections import deque


def flatten(nested):
    """Keeps flattening a nested lists of lists until  the
    first element of the resulting list is not a list.
    """
    if isinstance(nested, np.ndarray):
        return nested.ravel().tolist()
    else:
        while isinstance(nested[0], Iterable):
            nested = list(chain.from_iterable(nested))
        return nested


def format_permutations(array: Union[
        np.ndarray,
        List[List[int]],
        List[Tuple[int,...]],
        Tuple[Tuple[int,...],...],
        Tuple[List[int],...],
    ]) -> np.ndarray:
    """Permutations of inflation indices must leave the integers 0,
    corresponding to sources not being measured by the operator, invariant.
    In order to achieve this, this function shifts a permutation of sources
    by 1 and prepends it with the integer 0.

    Parameters
    ----------
    array : numpy.ndarray
        2-d array where each row is a permutations.

    Returns
    -------
    numpy.ndarray
        The processed list of permutations.
    """
    source_permutation = np.asarray(array) + 1
    return np.pad(source_permutation, ((0, 0), (1, 0)))


def clean_coefficients(cert: Dict[str, float],
                       chop_tol: float = 1e-10,
                       round_decimals: int = 3) -> Dict:
    """Clean the list of coefficients in a certificate.

    Parameters
    ----------
    cert : Dict[str, float]
      A dictionary containing as keys the monomials associated to the elements
      of the certificate and as values the corresponding coefficients.
    chop_tol : float, optional
      Coefficients in the dual certificate smaller in absolute value are
      set to zero. Defaults to ``1e-10``.
    round_decimals : int, optional
      Coefficients that are not set to zero are rounded to the number
      of decimals specified. Defaults to ``3``.

    Returns
    -------
    np.ndarray
      The cleaned-up coefficients.
    """
    if not cert:
        return cert
    chop_tol = np.abs(chop_tol)
    coeffs = np.array(list(cert.values()))
    good_coeffs_places = (np.abs(coeffs) > chop_tol)
    new_coeffs = np.compress(good_coeffs_places, coeffs)
    new_keys = [k for i,k in enumerate(cert.keys()) if good_coeffs_places[i]]
    if chop_tol > 0:
        # Try to take the smallest nonzero one and make it 1, when possible
        normalising_factor = np.min(np.abs(new_coeffs))
    else:
        # Take the largest nonzero one and make it 1
        normalising_factor = np.max(np.abs(new_coeffs))
    new_coeffs /= normalising_factor
    # Round
    new_coeffs = np.round(new_coeffs, decimals=round_decimals)
    return dict(zip(new_keys, new_coeffs.flat))


def eprint(*args, **kwargs):
    print(*args, file=stderr, **kwargs)


def partsextractor(thing_to_take_parts_of, indices) -> Tuple[int,...]:
    if hasattr(indices, '__iter__'):
        if len(indices) == 0:
            return tuple()
        elif len(indices) == 1:
            return (itemgetter(*indices)(thing_to_take_parts_of),)
        else:
            return itemgetter(*indices)(thing_to_take_parts_of)
    else:
        return itemgetter(indices)(thing_to_take_parts_of)


def expand_sparse_vec(sparse_vec: sps.coo_array,
                      conversion_style: str = "eq") -> sps.coo_array:
    """Expand a one-dimensional sparse matrix to its full form. Used to expand
    the solver arguments known_vars, lower_bounds, and upper_bounds."""
    assert conversion_style in {"eq", "lb", "ub"}, \
        "Conversion style must be `lb`, `ub`, or `eq`."
    nof_rows = sparse_vec.nnz
    nof_cols = sparse_vec.shape[1]
    if conversion_style == "eq":
        # Data values do not appear in '1' monomial column
        row = np.arange(nof_rows)
        col = sparse_vec.col
        data = np.ones(nof_rows)
    else:
        # Data values appear in '1' monomial column
        # Upper bound format: x <= a -> a - x >= 0
        row = np.repeat(np.arange(nof_rows), 2)
        col = np.vstack((
            sparse_vec.col,
            np.zeros(nof_rows)  # Assumes '1' monomial is first column
        )).T.ravel()
        data = np.vstack((
            -np.ones(nof_rows),
            sparse_vec.data,
        )).T.ravel()
    if conversion_style == "lb":
        # Lower bound format: x >= a -> x - a >= 0
        data = -data
    return sps.coo_array((data, (row, col)), shape=(nof_rows, nof_cols))


def vstack(blocks: tuple, format: str = 'coo') -> sps.coo_array:
    """Stack sparse matrices in coo_array form more efficiently."""
    non_empty = tuple(mat for mat in blocks if mat.shape[0])
    nof_blocks = len(non_empty)
    if nof_blocks > 1:
        if all(isinstance(block, sps.coo_array) for block in blocks):
            nof_rows = 0
            nof_cols = 0
            adjusted_rows = []
            for block in blocks:
                adjusted_rows.append(block.row + nof_rows)
                (block_len, block_wid) = block.shape
                nof_rows += block_len
                nof_cols = max(nof_cols, block_wid)
            mat_row = np.hstack(adjusted_rows)
            mat_col = np.hstack(tuple(block.col for block in blocks))
            mat_data = np.hstack(tuple(block.data for block in blocks))
            return sps.coo_array((mat_data, (mat_row, mat_col)),
                                  shape=(nof_rows, nof_cols)).asformat(format)
        else:
            return sps.vstack(blocks, format)
    elif nof_blocks == 1:
        return non_empty[0]
    else:
        return sps.coo_array([])


def perm_combiner(old_perms: np.ndarray, new_perms: np.ndarray) -> np.ndarray:
    combined = np.take(old_perms, new_perms, axis=1)
    new_shape = (-1, new_perms.shape[1])
    return combined.reshape(new_shape)


def all_and_maximal_cliques(adjmat: np.ndarray,
                            max_n=0,
                            isolate_maximal=True) -> (List, List):
    """Based on NetworkX's `enumerate_all_cliques`.
    This version uses native Python sets instead of numpy arrays.
    (Performance comparison needed.)

    Parameters
    ----------
    adjmat : numpy.ndarray
      A boolean numpy array representing the adjacency matrix of an undirected graph.
    max_n : int, optional
      A cutoff for clique size reporting. Default 0, meaning no cutoff.
    isolate_maximal : bool, optional
      A flag to disable filtering for maximality, which can increase performance. True by default.

    Returns
    -------
    Tuple[List, List]
      A list of all cliques as well as a list of maximal cliques. The maximal cliques list will be empty if the
      `isolate_maximal` flag is set to False.
    """
    all_cliques = [[]]
    maximal_cliques = []
    verts = tuple(range(adjmat.shape[0]))
    initial_cliques = [[u] for u in verts]
    nbrs_mat = np.triu(adjmat, k=1)
    initial_cnbrs = [np.flatnonzero(nbrs_mat[u]).tolist() for u in verts]
    nbrs = list(map(set, initial_cnbrs))
    queue = deque(zip(initial_cliques, initial_cnbrs))
    there_is_a_cutoff = (max_n <= 0)
    while queue:
        base, cnbrs = queue.popleft()
        all_cliques.append(base)
        if isolate_maximal and not len(cnbrs):
            base_as_set = set(base)
            if not any(base_as_set.issubset(superbase) for (superbase, _) in queue):
                maximal_cliques.append(base)
        elif there_is_a_cutoff or len(base) < max_n:
            for i, u in enumerate(cnbrs):
                new_base = base.copy()
                new_base.append(u)
                new_cnbrs = list(filter(nbrs[u].__contains__, cnbrs[i+1:]))
                queue.append((new_base, new_cnbrs))
    return all_cliques, maximal_cliques


def make_numerical(symbolic_expressions: Dict[Any, sympy.core.expr.Expr],
                   symbols_to_values: Dict[sympy.core.symbol.Symbol, float]
                   ) -> Dict[Any, float]:
    """Replace the symbols in the values of a dictionary by the corresponding
    numerical values.
    Parameters
    ----------
    symbolic_expressions : Dict[Any, sympy.core.expr.Expr]
        Dictionary where the values are symbolic expressions of some variables.
    symbols_to_values : Dict[sympy.core.symbol.Symbol, float]
        Correspondence of the variables in the expressions and their associated
        numerical values.
    Returns
    -------
    Dict[Any, float]
        The dictionary with same keys and evaluated expressions as values.
    """
    numeric_values = {}
    for k, v in symbolic_expressions.items():
        try:
            numeric_values[k] = float(v.evalf(subs=symbols_to_values))
        except AttributeError:
            numeric_values[k] = float(v)
    return numeric_values



"""
This file contains helper functions to write and export the problems to various
formats.

@authors: Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens
"""
import numpy as np

from copy import deepcopy
from scipy.io import savemat
from warnings import warn


def convert_to_human_readable(problem):
    """Convert the SDP relaxation to a human-readable format.

    Parameters
    ----------
    problem : :class:`inflation.InflationSDP`
        The SDP relaxation to write.

    Returns
    -------
    Tuple[str, numpy.ndarray, List, List]
        The first element is the objective function in a string, the second is
        a matrix of strings as the symbolic representation of the moment
        matrix, the third is a list of variable upper and lower bounds, and the
        fourth is a list of equality constraints of the form ``line = 0``.
    """
    matrix = deepcopy(problem.momentmatrix).astype(object)
    ### Process moment matrix
    # Replacer for constants
    constants = {moment.idx: str(value)
                 for moment, value in problem.known_moments.items()}
    # Replacer for semiknowns
    semiknowns = {}
    for key, val in problem.semiknown_moments.items():
        val_str = val[1].name.replace(", ", ";")
        semiknowns[key.idx] = f"{val[0]}*{val_str}"
    semiknown_replacer = np.vectorize(lambda x: semiknowns.get(x, str(x)))
    # Replacer for remaining symbols
    monomials = {}
    for mon in problem.moments:
        monomials[mon.idx] = mon.name.replace(", ", ";")

    def replace_known(monom):
        try:
            replacement = monomials.get(float(monom), monom)
        except ValueError:
            replacement = monom
        return replacement
    known_replacer = np.vectorize(replace_known)
    for ii, row in enumerate(matrix):
        for jj, col in enumerate(row):
            matrix[ii, jj] = constants.get(col, col)
    matrix = semiknown_replacer(matrix)
    matrix = np.triu(known_replacer(matrix).astype(object))

    ### Process objective
    is_first = True
    try:
        independent_term = float(problem.objective[problem.One])
        if abs(independent_term) > 1e-8:
            objective = str(independent_term)
            is_first = False
        else:
            objective = ""
    except KeyError:
        objective = ""
    for variable, coeff in problem.objective.items():
        if variable != problem.One:
            if (coeff < 0) or is_first:
                objective += f"{float(coeff)}*{variable.name}"
            else:
                objective += f"+{float(coeff)}*{variable.name}"
            is_first = False

    ### Process bounds
    bounded_vars = sorted(list(set(problem.moment_upperbounds.keys()).union(
                               set(problem.moment_lowerbounds.keys()))),
                          key=lambda x: x.name)
    bounds = np.zeros((len(bounded_vars), 3), dtype=object)
    for idx, var in enumerate(bounded_vars):
        bounds[idx, 0] = var
        bounds[idx, 1] = problem.moment_lowerbounds.get(var, None)
        bounds[idx, 2] = problem.moment_upperbounds.get(var, None)

    ### Process equalities
    equalities = []
    for eq in problem.moment_equalities:
        equality = ""
        for monom, coeff in eq.items():
            equality += "+" if coeff > 0 else "-"
            if monom == problem.One:
                equality += str(abs(coeff))
            else:
                if np.isclose(abs(coeff), 1):
                    equality += monom.name
                else:
                    equality += f"{abs(coeff)}*{monom.name}"
        equalities.append(equality[1:] if equality[0] == "+" else equality)
    
    ### Process inequalities
    inequalities = []
    for ineq in problem.moment_inequalities:
        inequality = ""
        for monom, coeff in ineq.items():
            inequality += "+" if coeff > 0 else "-"
            if monom == problem.One:
                inequality += str(abs(coeff))
            else:
                if np.isclose(abs(coeff), 1):
                    inequality += monom.name
                else:
                    inequality += f"{abs(coeff)}*{monom.name}"
        inequalities.append(inequality[1:] if inequality[0] == "+"
                            else inequality)
    return objective, matrix, bounds.tolist(), equalities, inequalities


def write_to_csv(problem, filename):
    """
    Export the problem in a human-readable form in a CSV table.

    Parameters
    ----------
    problem : :class:`inflation.InflationSDP`
        The SDP relaxation to write.
    filename : str
        The name of the file to write the CSV table to.
    """
    objective, matrix, bounds, equalities, inequalities \
        = convert_to_human_readable(problem)
    f = open(filename, "w")
    f.write("Objective: " + objective + "\n")
    for matrix_line in matrix:
        f.write(
            str(list(matrix_line))[1:-1].replace(" ", "").replace("\'", ""))
        f.write("\n")
    f.write("Bounds:\n")
    f.write("Variable,lower,upper\n")
    for bound_line in bounds:
        f.write(str(bound_line)[1:-1].replace(" ", ""))
        f.write("\n")
    f.write("\nEqualities (format: line = 0):\n")
    for equality in equalities:
        f.write(equality)
        f.write("\n")
    f.write("\nInequalities (format: line >= 0):\n")
    for inequality in inequalities:
        f.write(inequality)
        f.write("\n")
    f.close()


def write_to_mat(problem, filename):
    """
    Export the problem to MATLAB .mat file.

    Parameters
    ----------
    problem : inflation.InflationSDP
        The SDP relaxation to write.
    filename : str
        The filename of the MATLAB .mat file to be created.

    Notes
    -----
    MATLAB does not like 0s, so we shift all values by 1 if the variable 0
    exists.

    The exported MATLAB .mat file contains the following variables:
    - moments_idx2name : numpy.ndarray
        An array containing the indices and names of the moments.
    - momentmatrix : numpy.ndarray
        The moment matrix with the offset applied.
    - objective : list
        A list of lists containing the indices and coefficients of the objective
        moments.
    - known_moments : list
        A list of lists containing the indices and values of the known moments.
    - semiknown_moments : numpy.ndarray
        A 2D array containing the indices, factors, and final indices of the
        semi-known moments.
    - moment_lowerbounds : list
        A list of lists containing the indices and lower bounds of the moments.
    - moment_upperbounds : list
        A list of lists containing the indices and upper bounds of the moments.
    - moment_equalities : list
        A list of dictionaries representing moment equalities, where each
        dictionary contains the indices and coefficients of the moments involved
        in the equality.
    - moment_inequalities : list
        A list of dictionaries representing moment inequalities, where each
        dictionary contains the indices and coefficients of the moments involved
        in the inequality.
    """
    # MATLAB does not like 0s, so we shift all by 1 if the variable 0 exists
    offset = 1 if problem.momentmatrix_has_a_zero else 0
    final_positions_matrix = problem.momentmatrix + offset
    known_moments = [[mon.idx + offset, val]
                     for mon, val in problem.known_moments.items()]
    semiknown_initial = []
    semiknown_factors = []
    semiknown_final   = []
    for semiknown in problem.semiknown_moments.items():
        semiknown_initial.append(semiknown[0].idx + offset)
        semiknown_factors.append(semiknown[1][0])
        semiknown_final.append(semiknown[1][1].idx + offset)
    semiknown_moments = np.vstack([semiknown_initial,
                                   semiknown_factors,
                                   semiknown_final]).T
    objective   = [[mon.idx + offset, float(coeff)]
                   for mon, coeff in problem.objective.items()
                   if abs(coeff) > 1e-8]
    lowerbounds = [[mon.idx + offset, bnd]
                   for mon, bnd in problem.moment_lowerbounds.items()]
    upperbounds = [[mon.idx + offset, bnd]
                   for mon, bnd in problem.moment_upperbounds.items()]
    names       = np.array([[mon.idx + offset, mon.name]
                            for mon in problem.moments], dtype=object)
    equalities  = []
    for eq in problem.moment_equalities:
        equality = {'moments': np.array([mon.idx + offset for mon in eq]),
                    'coeffs':  np.array(list(eq.values()))}
        equalities.append(equality)
  
    inequalities  = []
    for ineq in problem.moment_inequalities:
        inequality = {'moments': np.array([mon.idx + offset for mon in ineq]),
                      'coeffs':  np.array(list(ineq.values()))}
        inequalities.append(inequality)

    savemat(filename,
            mdict={"moments_idx2name":    names,
                   "momentmatrix":        final_positions_matrix,
                   "objective":           objective,
                   "known_moments":       known_moments,
                   "semiknown_moments":   semiknown_moments,
                   "moment_lowerbounds":  lowerbounds,
                   "moment_upperbounds":  upperbounds,
                   "moment_equalities":   equalities,
                   "moment_inequalities": inequalities
                   }
            )


def write_to_sdpa(problem, filename):
    """
    Export the problem to a file in .dat-s format, per as
    http://euler.nmt.edu/~brian/sdplib/FORMAT

    Parameters
    ----------
    problem : inflation.InflationSDP
        The SDP relaxation to write.
    filename : str
        The filename to write the SDP problem to.

    Notes
    -----
    This function exports the given SDP relaxation problem to a file in the
    .dat-s format. The .dat-s format is a specific format used by the SDPA
    software for semidefinite programming.

    For more information about the .dat-s format, see the specifications at:
    http://euler.nmt.edu/~brian/sdplib/FORMAT
    """
    # Compute actual number of variables: all in the moment matrix, minus those
    # with known values, minus those that participate in LPI constraints, plus
    # those where the unknown part of the LPI constraint is not in the original
    # moment matrix
    potential_nvars = problem.momentmatrix.max() - 1
    if len(problem.known_moments) == 1:
        known_vars = 0
    else:
        known_vars = (len(problem.known_moments)
                      - problem.momentmatrix_has_a_zero
                      - problem.momentmatrix_has_a_one)
    semiknown_vars = len(problem.semiknown_moments)
    new_vars       = sum([mon[1][1].idx > problem.momentmatrix.max()
                          for mon in problem.semiknown_moments.items()])
    nvars = potential_nvars - known_vars - semiknown_vars + new_vars

    known_moments_indices = {mon.idx: coeff
                             for mon, coeff in problem.known_moments.items()}
    semiknowns  = {mon.idx: (subs[0], subs[1].idx)
                   for mon, subs in problem.semiknown_moments.items()}
    lines       = []
    var_corresp = {}
    new_var     = 1
    block       = 1
    blockstruct = [str(problem.momentmatrix.shape[0])]
    for ii, row in enumerate(problem.momentmatrix):
        for jj, var in enumerate(row):
            if jj >= ii:
                if var == 0:
                    pass
                elif var == 1:
                    lines.append(f"0\t{block}\t{ii+1}\t{jj+1}\t-1.0\n")
                elif var in known_moments_indices.keys():
                    coeff = known_moments_indices[var]
                    lines.append(
                        f"0\t{block}\t{ii+1}\t{jj+1}\t-{abs(coeff)}\n")
                elif var in semiknowns.keys():
                    coeff, subs = semiknowns[var]
                    try:
                        var = var_corresp[subs]
                        lines.append(
                            f"{var}\t{block}\t{ii+1}\t{jj+1}\t{coeff}\n")
                    except KeyError:
                        var_corresp[subs] = new_var
                        lines.append(
                            f"{new_var}\t{block}\t{ii+1}\t{jj+1}\t{coeff}\n")
                        new_var += 1
                else:
                    try:
                        var = var_corresp[var]
                        lines.append(f"{var}\t{block}\t{ii+1}\t{jj+1}\t1.0\n")
                    except KeyError:
                        var_corresp[int(var)] = new_var
                        lines.append(
                            f"{new_var}\t{block}\t{ii+1}\t{jj+1}\t1.0\n")
                        new_var += 1

    # Prepare objective
    objective = np.zeros(nvars)
    for variable, coeff in problem.objective.items():
        if variable == problem.One:
            if abs(coeff) > 1e-8:
                warn(f"Export removed the constant {coeff} from the objective")
        else:
            objective[var_corresp[variable.idx] - 1] = coeff
    objective = str(objective.tolist()).replace("[", "").replace("]", "")
    objective_constant = problem.objective[problem.One]

    # Prepare upper bounds
    if len(problem.moment_upperbounds) > 0:
        block += 1
        ii     = 1
        block_size = len([mon for mon in problem.moment_upperbounds.keys()
                          if ((mon != problem.One)
                          and (mon not in problem.known_moments.keys())
                          and (mon not in problem.semiknown_moments.keys()))])
        blockstruct.append(str(-block_size))
    for var, ub in problem.moment_upperbounds.items():
        if ((var != problem.One)
            and (var not in problem.known_moments.keys())
            and (var not in problem.semiknown_moments.keys())):
            var = var_corresp[var.idx]
            lines.append(f"{var}\t{block}\t{ii}\t{ii}\t-1.0\n")
            if abs(ub) > 1e-8:
                lines.append(f"0\t{block}\t{ii}\t{ii}\t-{ub}\n")
            ii += 1

    # Prepare lower bounds
    if len(problem.moment_lowerbounds) > 0:
        block += 1
        ii     = 1
        block_size = len([mon for mon in problem.moment_lowerbounds.keys()
                          if ((mon != problem.One)
                          and (mon not in problem.known_moments.keys())
                          and (mon not in problem.semiknown_moments.keys()))])
        blockstruct.append(str(-block_size))
    for var, lb in problem.moment_lowerbounds.items():
        if ((var != problem.One)
            and (var not in problem.known_moments.keys())
            and (var not in problem.semiknown_moments.keys())):
            var = var_corresp[var.idx]
            lines.append(f"{var}\t{block}\t{ii}\t{ii}\t1.0\n")
            if abs(lb) > 1e-8:
                lines.append(f"0\t{block}\t{ii}\t{ii}\t{lb}\n")
            ii += 1

    # Prepare equalities
    if len(problem.moment_equalities) > 0:
        block += 1
        ii     = 1
        block_size = 2*len(problem.moment_equalities)
        blockstruct.append(str(-block_size))
    for equality in problem.moment_equalities:
        for var, coeff in equality.items():
            if problem.known_moments.get(var, None) not in [0]:
                if var != problem.One:
                    var = var_corresp[var.idx]
                    lines.append(f"{var}\t{block}\t{ii}\t{ii}\t{coeff}\n")
                    lines.append(f"{var}\t{block}\t{ii+1}\t{ii+1}\t{-coeff}\n")
                else:
                    lines.append(f"0\t{block}\t{ii}\t{ii}\t{-coeff}\n")
                    lines.append(f"0\t{block}\t{ii+1}\t{ii+1}\t{coeff}\n")
        ii += 2

    # Prepare inequalities
    if len(problem.moment_inequalities) > 0:
        block += 1
        ii = 1
        block_size = len(problem.moment_inequalities)
        blockstruct.append(str(-block_size))
    for inequality in problem.moment_inequalities:
        sumcoeffs = 0
        for var, coeff in inequality.items():
            if problem.known_moments.get(var, None) not in [0]:
                if var != problem.One:
                    var = var_corresp[var.idx]
                    lines.append(f"{var}\t{block}\t{ii}\t{ii}\t{coeff}\n")
                else:
                    sumcoeffs += coeff
        if sumcoeffs != 0:
            lines.append(f"0\t{block}\t{ii}\t{ii}\t{-sumcoeffs}\n")
        ii += 2
    file_ = open(filename, "w")
    file_.write("\"file " + filename + " generated by inflation\"\n")
    if abs(objective_constant) > 1e-8:
        file_.write("\"the objective contains a constant term of value "
                    + str(objective_constant) + " that is not included in the "
                    + "program below\"\n")
    file_.write(f"{nvars} = number of vars\n")
    file_.write(f"{block} = number of blocks\n")
    file_.write("(" + ",".join(blockstruct) + ") = BlockStructure\n")
    file_.write("{" + objective + "}\n")
    for line in sorted(lines, key=lambda x: (int(x.split("\t")[1]),
                                             int(x.split("\t")[0]))):
        file_.write(line)
    file_.close()



"""
This file contains helper functions to which can be accelerated by JIT
compilation in numba.
@authors: Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens
"""
import numpy as np

try:
    from numba import jit, bool_
    from numba.types import int_
    nopython = True
except ImportError:
    print("Warning: NOT using Numba for JIT compilation.")
    def jit(*args, **kwargs):
        return lambda f: f
    bool_    = bool
    int_     = int
    nopython = False
    prange   = range
cache    = True
if not nopython:
    bool_  = bool
    int_ = int

@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_mon_to_lexrepr_bool(mon: np.ndarray,
                           lexorder: np.ndarray) -> np.array:
    """Convert a monomial to its lexicographic representation, in the form of
    an array of integers representing the lexicographic rank of each operator.

    Parameters
    ----------
    mon : numpy.ndarray
        Input monomial in 2d array format.
    lexorder : numpy.ndarray
        Matrix with rows as operators where the index of the row gives
        the lexicographic order of the operator.

    Returns
    -------
    np.array
        Boolean vector of length lexorder, with 1 if something in mon is in
        lexorder
    """
    length = lexorder.shape[0]
    in_lex = np.zeros(length, dtype=bool_)
    for i in range(length):
        standard_op = lexorder[i]
        for op in mon:
            if np.array_equal(op, standard_op):
                in_lex[i] = True
                break
    return in_lex

@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_outer_bitwise_or(a: np.ndarray, b: np.ndarray):
    a_adj = np.expand_dims(a, 1)
    b_adj = b.reshape((1,)+b.shape)
    temp = np.logical_or(a_adj, b_adj)
    return temp.reshape((-1, *temp.shape[2:]))

@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_outer_bitwise_xor(a: np.ndarray, b: np.ndarray):
    a_adj = np.expand_dims(a, 1)
    b_adj = b.reshape((1,)+b.shape)
    temp = np.logical_xor(a_adj, b_adj)
    return temp.reshape((-1, *temp.shape[2:]))

@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_is_do_conditional(monomial: np.ndarray) -> bool_:
    """Determine whether a given atomic monomial admits an identification with
    a do conditional of the original scenario.

    Parameters
    ----------
    monomial : np.ndarray
        List of operators, denoted each by a list of indices

    Returns
    -------
    bool
        Whether the monomial is knowable or not.
    """
    if len(monomial) <= 1:
        return True
    # Mappable monomials have at most one copy of each source in the DAG
    for source in monomial.T[1:-2]:
        if len(np.unique(source[np.flatnonzero(source)])) > 1:
            return False
    return True


import unittest
import warnings
import numpy as np

from inflation import InflationProblem, InflationLP
from inflation.symmetry_utils import (discover_distribution_symmetries,
                                      group_elements_from_generators)


class TestSymmetry(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore", category=UserWarning)

    PR_box = np.zeros((2, 2, 2, 2))
    for a,b,x,y in np.ndindex(*PR_box.shape):
        if np.bitwise_xor(a,b) == np.bitwise_and(x,y):
            PR_box[a,b,x,y] = 0.5

    GHZ = np.zeros((2, 2, 2, 1, 1, 1))
    GHZ[0, 0, 0, 0, 0, 0] = 1/2
    GHZ[1, 1, 1, 0, 0, 0] = 1/2

    bellScenario = InflationProblem({"Lambda": ["A", "B"]},
                                    outcomes_per_party=[2, 2],
                                    settings_per_party=[2, 2],
                                    inflation_level_per_source=[1],
                                    classical_sources='all')

    triangle = InflationProblem({"Lambda": ["A", "B"],
                                 "Mu": ["B", "C"],
                                 "Sigma": ["C", "A"]},
                                outcomes_per_party=[2, 2, 2],
                                inflation_level_per_source=[2, 1, 1])

    PRbox_symmetries = discover_distribution_symmetries(PR_box,
                                                        bellScenario)

    def test_discover(self):
        # Order: (a=0,x=0), (a=1,x=0), (a=0,x=1), (a=1,x=1),
        #        (b=0,y=0), (b=1,y=0), (b=0,y=1), (b=1,y=1)
        # There's five symmetries: flip parties, flip X, flip Y, and flip the
        # outcomes of each. Out of all these, the only valid ones are those that
        # do not change a+b+xy mod 2
        # Identity
        symmetries = [[0, 1, 2, 3, 4, 5, 6, 7]]
        # Flip outcomes in x=1, and flip y
        symmetries += [[0, 1, 3, 2, 6, 7, 4, 5]]
        # Flip outcomes in x=0, flip b and y
        symmetries += [[1, 0, 2, 3, 7, 6, 5, 4]]
        # Flip a and b
        symmetries += [[1, 0, 3, 2, 5, 4, 7, 6]]
        # Flip x, flip b outcomes in y=1
        symmetries += [[2, 3, 0, 1, 4, 5, 7, 6]]
        # Flip x, flip y, flip a in x=1, flip b in y=0
        symmetries += [[2, 3, 1, 0, 7, 6, 4, 5]]
        # Flip x, flip y, flip a in x=0, flip b in y=1
        symmetries += [[3, 2, 0, 1, 6, 7, 5, 4]]
        # Flip x, flip a, flip b in y=0
        symmetries += [[3, 2, 1, 0, 5, 4, 6, 7]]
        # All the above, but swapping A and B
        swapped = [symm[4:] + symm[:4] for symm in symmetries]
        symmetries = symmetries + swapped
        self.assertSetEqual(set(map(tuple, symmetries)),
                            set(map(tuple, self.PRbox_symmetries)),
                            "Failed to discover the symmetries of the PR box.")

    def test_discover_inflation(self):
        GHZ_symmetries = discover_distribution_symmetries(self.GHZ,
                                                          self.triangle)
        # Order: a11=0 a11=1 a12=0 a12=1 b11=0 b11=1 b21=0 b21=1 c11=0 c11=1
        # There's only three symmetries: swap all outcomes, swap A and B, and
        # both at the same time
        # Identity
        symmetries = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]
        # Flip all outcomes
        symmetries += [[1, 0, 3, 2, 5, 4, 7, 6, 9, 8]]
        # Flip A and B
        symmetries += [[4, 5, 6, 7, 0, 1, 2, 3, 8, 9]]
        # Both at the same time
        symmetries += [[5, 4, 7, 6, 1, 0, 3, 2, 9, 8]]
        self.assertSetEqual(set(map(tuple, symmetries)),
                            set(map(tuple, GHZ_symmetries)),
                            "Failed to discover the symmetries of the PR box.")

    def test_group_elements_from_generators(self):
        generators = [[1, 2, 0], [1, 0, 2]]
        elements = group_elements_from_generators(generators)
        truth = [[0, 1, 2], [1, 0, 2], [2, 1, 0],
                 [2, 0, 1],  [0, 2, 1], [1, 2, 0]]
        self.assertSetEqual(set(map(tuple, elements)),
                            set(map(tuple, truth)),
                            "Failed to generate S3 from generators.")

    def test_desymmetrized_certificate(self):
        self.bellScenario.add_symmetries(self.PRbox_symmetries)
        lp = InflationLP(self.bellScenario, verbose=0)
        lp.set_distribution(self.PR_box)
        lp.solve(default_non_negative=False, solve_dual=False)
        certificate = lp.desymmetrize_certificate()
        truth = {'P[A_0=0]': 0.25, 'P[A_0=1]': 0.25, 'P[A_1=1]': 0.25, 'P[A_1=0]': 0.25,
                 'P[B_0=0]': 0.25, 'P[B_0=1]': 0.25, 'P[B_1=0]': 0.25, 'P[B_1=1]': 0.25,
                 'P[A_0=0 B_0=0]': -0.375, 'P[A_0=0 B_1=0]': -0.375, 'P[A_0=1 B_0=1]': -0.375, 'P[A_0=1 B_1=1]': -0.375,
                 'P[A_1=0 B_0=0]': -0.375, 'P[A_1=0 B_1=1]': -0.375, 'P[A_1=1 B_1=0]': -0.375, 'P[A_1=1 B_0=1]': -0.375,
                 'P[A_0=0 B_0=1]': 0.125, 'P[A_0=0 B_1=1]': 0.125, 'P[A_0=1 B_0=0]': 0.125, 'P[A_1=0 B_0=1]': 0.125,
                 'P[A_1=0 B_1=0]': 0.125, 'P[A_1=1 B_1=1]': 0.125, 'P[A_1=1 B_0=0]': 0.125, 'P[A_0=1 B_1=0]': 0.125
                 }
        self.assertDictEqual(certificate, truth,
                             "Failed to desymmetrize the CHSH inequality.")




import unittest
import numpy as np
import warnings

from sympy import Symbol

from inflation import InflationProblem, InflationSDP, max_within_feasible


class TestOptimize(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore", category=UserWarning)

    def PRbox(v):
        p = np.zeros((2, 2, 2, 2), dtype=object)
        for a, b, x, y in np.ndindex((2, 2, 2, 2)):
            p[a, b, x, y] = 1/4 * (1 + v*(-1)**(a + b + x * y))
        return p

    precision = 1e-5
    bellScenario = InflationProblem({"Lambda": ["A", "B"]},
                                    outcomes_per_party=[2, 2],
                                    settings_per_party=[2, 2],
                                    inflation_level_per_source=[1])
    sdp = InflationSDP(bellScenario)
    sdp.generate_relaxation("npa1")
    sdp.set_distribution(PRbox(Symbol("v")))
    symbolic_values = sdp.known_moments

    def test_bisect(self):
        v_crit = max_within_feasible(self.sdp,
                                     self.symbolic_values,
                                     "bisection",
                                     precision=self.precision)
        self.assertTrue(np.isclose(v_crit, 1/np.sqrt(2), self.precision),
                        "Bisection of the quantum critical visibility for the "
                        + "PR box is not achieving 1/sqrt(2).")

    def test_dual(self):
        v_crit = max_within_feasible(self.sdp, self.symbolic_values, "dual",
                                     precision=self.precision)
        self.assertTrue(np.isclose(v_crit, 1/np.sqrt(2), self.precision),
                        "Dual optimization of the quantum critical visibility "
                        + "for the PR box is not achieving 1/sqrt(2).")



"""
This file contains auxiliary functions for discovering symmetries

@authors: Emanuel-Cristian Boghiu, Elie Wolfe and Alejandro Pozas-Kerstjens
"""

from typing import List, Union

import numpy as np

from sympy.combinatorics import Permutation, PermutationGroup
from tqdm import tqdm
from . import InflationProblem

def discover_distribution_symmetries(distribution: np.ndarray,
                                     scenario: InflationProblem
                                     ) -> np.ndarray:
    """
    Given a distribution, find the symmetries of the distribution that are
    compatible with the symmetries of the scenario. The symmetries are
    represented as permutations of the lexicographic order of the events in
    the scenario.

    Parameters
    ----------
    distribution : numpy.ndarray
        The distribution to be analyzed. It must be a numpy array of shape
        ``[o1, ..., oN, s1, ..., sN]``, where ``oi`` and ``si`` are, respectively,
        the number of outcomes and settings for party ``i``.
    scenario : InflationProblem
        The scenario object.

    Returns
    -------
    numpy.ndarray
        The symmetries of the distribution that are compatible with the
        symmetries of the scenario. The symmetries are represented as
        permutations of the lexicographic order of the events in the
        scenario.
    """
    # Sanity checks
    parties = scenario.nr_parties
    if not isinstance(distribution, np.ndarray):
        raise TypeError("The distribution must be encoded in a numpy array.")
    if len(distribution.shape) != 2*parties:
        raise ValueError("The distribution must be encoded as an " + 
                         "2*nr_parties-dimensional array.")
    if np.any(distribution.shape[:parties] != scenario.outcomes_per_party):
        raise ValueError("The number of outcomes of the distribution and of " +
                         "the scenario do not match.")
    if np.any(distribution.shape[parties:] != scenario.settings_per_party):
        raise ValueError("The number of settings of the distribution and of " +
                         "the scenario do not match.")
    if not np.all(distribution >= 0):
        raise ValueError("The distribution contains negative values.")
    if not np.allclose(distribution.sum(axis=tuple(range(parties))), 1):
        raise ValueError("The distribution is not normalized for each setting.")

    original_dag_events_order = {tuple(op): i
                           for i, op in enumerate(scenario.original_dag_events)}

    ## Find the symmetries of the distribution.
    # First, build the knowable monomials of maximum length of the original
    # DAG, and their values under the distribution.
    original_dag_monomials_values = {}
    original_dag_monomials_lexboolvecs = []
    for ins in np.ndindex(*scenario.settings_per_party):
        for outs in np.ndindex(*scenario.outcomes_per_party):
            # Build the monomial corresponding to p(outs...|ins...)
            original_dag_lexboolvec = np.zeros(len(scenario.original_dag_events),
                                               dtype=bool)
            for p, (x, a) in enumerate(zip(ins, outs)):
                original_dag_lexboolvec[
                    original_dag_events_order[(p + 1, x, a)]] = True
            original_dag_monomials_lexboolvecs += [original_dag_lexboolvec]
            # Calculate its value under the distribution
            _value = distribution[(*outs, *ins)]
            _hash = original_dag_lexboolvec.tobytes()
            original_dag_monomials_values[_hash] = _value
    original_dag_monomials_lexboolvecs = np.array(
                                             original_dag_monomials_lexboolvecs)
    original_values_1d = np.array([original_dag_monomials_values[mon.tobytes()]
                                 for mon in original_dag_monomials_lexboolvecs])
    good_perms  = []
    for perm_lexorder in tqdm(scenario._all_possible_symmetries,
                              desc="Discovering distribution symmetries",
                              disable=not scenario.verbose):
        perm_orig   = lexperm_to_origperm(perm_lexorder, scenario)
        lexboolvecs = original_dag_monomials_lexboolvecs.copy()
        lexboolvecs = lexboolvecs[:, perm_orig]  # permute the columns
        new_values_1d = np.array([original_dag_monomials_values[mon.tobytes()]
                                    for mon in lexboolvecs])
        if np.allclose(new_values_1d, original_values_1d):
            good_perms += [perm_lexorder]
    if scenario.verbose > 0:
        print(f"Found {len(good_perms)} symmetries.")

    return np.array(good_perms)

def group_elements_from_generators(generators: Union[np.ndarray,
                                                List[np.ndarray]]
                                    ) -> np.ndarray:
    """
    Given a set of generators of some permutation group, return the group
    elements in lexicographic order.

    Parameters
    ----------
    generators : Union[numpy.ndarray, List[numpy.ndarray]]
        The generators of the permutation group. Each generator is a permutation
        of the indices ``[0...n]``.

    Returns
    -------
    numpy.ndarray
        The elements of the symmetry group as permutatiosn of th lexicographic
        order.
    """
    G = PermutationGroup([Permutation(perm)
                          for perm in np.unique(generators, axis=0)])
    group_elements = np.array(list(G.generate_schreier_sims(af=True)))
    return group_elements[np.lexsort(np.rot90(group_elements))]

def interpret_lexorder_symmetry(perm: np.ndarray,
                                scenario: InflationProblem) -> dict:
    """Gives a human-readable form of a permutation of the lexicographic order
    of the events in the scenario.

    Parameters
    ----------
    perm : numpy.ndarray
        The permutation of the lexicographic order of the events in the
        scenario.
    scenario : InflationProblem
        The scenario object.

    Returns
    -------
    dict
        A dictionary that maps the names of the events in lexicographic order
        to the names of the corresponding events under the symmetry.
    """
    return dict(zip(scenario._lexrepr_to_names,
                    scenario._lexrepr_to_names[perm]))

def interpret_original_symmetry(perm: np.ndarray,
                                scenario: InflationProblem) -> dict:
    """Gives a human-readable form of a permutation of the observable events in
    the scenario.

    Parameters
    ----------
    perm : numpy.ndarray
        The permutation of the events in the scenario.
    scenario : InflationProblem
        The scenario object.

    Returns
    -------
    dict
        A dictionary that maps the names of the events in lexicographic order
        to the names of the corresponding events under the symmetry.
    """
    return dict(zip(scenario._original_event_names,
                    scenario._original_event_names[perm]))

def lexperm_to_origperm(lexperm: np.ndarray,
                        scenario: InflationProblem) -> np.ndarray:
    """
    Given a permutation of the lexicographic order of the events in the
    scenario, return the permutation of the observable events in the scenario.

    Parameters
    ----------
    lexperm : numpy.ndarray
        The permutation of the lexicographic order of the events in the
        scenario.
    scenario : InflationProblem
        The scenario object.

    Returns
    -------
    numpy.ndarray
        The corresponding permutation of the observable events in the scenario.
    """
    return scenario._lexidx_to_origidx[lexperm][scenario._canonical_lexids]



"""
This file contains the functions to send the problems to SDP solvers.

@authors: Emanuel-Cristian Boghiu, Elie Wolfe and Alejandro Pozas-Kerstjens
"""
from __future__ import annotations
import numpy as np

from copy import deepcopy
from gc import collect
from scipy.sparse import eye, coo_array
from sys import stdout
from typing import List, Dict, Any, Iterator


def solveSDP_MosekFUSION(mask_matrices: Dict = None,
                         objective: Dict = None,
                         known_vars: Dict = None,
                         semiknown_vars: Dict = None,
                         inequalities: List[Dict] = None,
                         equalities: List[Dict] = None,
                         solve_dual: bool = True,
                         default_non_negative: bool = False,
                         feas_as_optim: bool = False,
                         verbose: int = 0,
                         solverparameters: Dict = {},
                         process_constraints: bool = True
                         ) -> Dict[str, Any]:
    r"""Internal function to solve the SDP with the `MOSEK Fusion API
    <https://docs.mosek.com/latest/pythonfusion/index.html>`_.

    Now follows an extended description of how the SDP is encoded. In general,
    it is preferred to solve using the dual formulation, which is the default.

    The primal is written as follows:

    .. math::

        \text{max}\quad & c_0 + c\cdot x\\
        \text{s.t.}\quad & F_0 + \sum F_i x_i \succeq 0

    :math:`F_0` is the constant entries of the moment matrix, and :math:`F_i`
    is the matrix whose entry :math:`(n,m)` stores the value of the coefficient
    of the moment :math:`x_i` at position :math:`(n,m)` in the moment matrix.

    The dual of the equation above is:

    .. math::

        \text{min}\quad & c_0+\text{Tr}(Z\cdot F_0)\\
        \text{s.t.}\quad & \text{Tr}(Z\cdot F_i) = - c_i \,\forall\, i,\\
        &Z \succeq 0.

    Typically, all the probability information is stored in :math:`F_0`, and
    the coefficients :math:`F_i` do not depend on the probabilities. However,
    if we use LPI constraints (see, e.g., `arXiv:2203.16543
    <http://www.arxiv.org/abs/2203.16543/>`_), then :math:`F_i` can depend on
    the probabilities. The form of the SDP does not change, in any case.

    If we have a constant objective function, then we have a feasibility
    problem. It can be rewritten into the following optimization problem:

    .. math::
        \text{max}\quad&\lambda\\
        \text{s.t.}\quad& F_0 + \sum F_i x_i - \lambda \cdot 1 \succeq 0,

    which achieves :math:`\lambda\geq 0` if the original problem is feasible
    and :math:`\lambda<0` otherwise. The dual of this problem is:

    .. math::
        \text{min}\quad & \text{Tr}(Z\cdot F_0) \\
        \text{s.t.}\quad & \text{Tr}(Z\cdot F_i) = 0 \,\forall\, i,\\
            & Z \succeq 0,\,\text{Tr} Z = 1.

    This still allows for the extraction of certificates. If we use a
    :math:`Z_{P_1}` obtained from running the problem above on the probability
    distribution :math:`P_1`, and we find that
    :math:`\text{Tr}[Z_{P_1}\cdot F_0(P_2)] < 0`, then clearly this is an upper
    bound of the optimal value of the problem, and thus we can certify that the
    optimisation will be negative when using :math:`P_2`.

    If we have upper and lower bounds on the variables, the problems change as
    follows:

    .. math::
        \text{max}\quad & c_0 + c\cdot x \\
        \text{s.t.}\quad & F_0 + \sum F_i x_i \succeq 0,\\
        & x_i - l_i \geq 0,\\
        & u_i - x_i \geq 0,

    with dual:

    .. math::
        \text{min}\quad & \text{Tr}(Z\cdot F_0 - L\cdot l + U\cdot u) \\
        \text{s.t.}\quad & \text{Tr}(Z \cdot F_i) = -c_i+U_i-L_i\,\forall\,i,\\
        & Z \succeq 0,\,L \geq 0,\,U \geq 0.

    The relaxed feasibility problems change accordingly.

    Parameters
    ----------
    mask_matrices : dict
        A dictionary with keys as monomials and values as scipy sparse arrays
        indicating the locations of the monomial in the moment matrix.
    objective : dict, optional
        Dictionary with keys as monomials and as values the monomial's
        coefficient in the objective function.
    known_vars : dict, optional
        Dictionary of values for monomials (keys).
    semiknown_vars : dict, optional
        Dictionary encoding proportionality constraints between
        different monomials.
    inequalities : list, optional
        List of inequalities encoded as dictionaries of coefficients.
    equalities : list, optional
        List of equalities encoded as dictionaries of coefficients.
    solve_dual : bool, optional
        Whether to solve the dual (True) or primal (False) formulation. By
        default ``True``.
    default_non_negative: bool, optional
        Whether to set default primal variables as non-negative (True) or not
        (False). By default, ``False``.
    feas_as_optim : bool, optional
        Whether to treat feasibility problems, where the objective is,
        constant, as an optimisation problem. By default ``False``.
    verbose : int, optional
        How much information to display to the user. By default ``0``.
    solverparameters : dict, optional
        Dictionary of parameters to pass to the MOSEK solver, see `MOSEK's
        documentation
        <https://docs.mosek.com/latest/pythonfusion/solver-parameters.html>`_.
    process_constraints: bool, optional
        Whether to remove the simple equalities constraints contained
        in the `semiknown_vars` arguments by eliminating variables (True)
        or pass them to the solver as equality constraints (False). By
        default ``True``.

    Returns
    -------
    Tuple[Dict, float, str]
        The first element of the tuple is a dictionary containing the
        optimisation information such as the 1) primal objective value,
        2) the moment matrix, 3) the dual values, 4) the certificate and
        a 5) dictionary of values for the monomials, in the following keys in
        the same order: 1) ``sol``, 2) ``G``, 3) ``Z``, 4)
        ``dual_certificate``, 5) ``xi``. The second element is the objective
        value and the last is the problem status.
    """
    import mosek
    from mosek.fusion import Matrix, Model, ObjectiveSense, Expr, Domain, \
        OptimizeError, SolutionError, \
        AccSolutionStatus, ProblemStatus

    def scipy_to_mosek(mat: coo_array) -> Matrix:
        internal_mat = mat.tocoo(copy=False)
        return Matrix.sparse(*internal_mat.shape,
                             np.asarray(internal_mat.row, dtype=np.int32),
                             np.asarray(internal_mat.col, dtype=np.int32),
                             internal_mat.data)
    def coo_getcol(mat: coo_array, i: int) -> coo_array:
        condition = (mat.col == i)
        new_col = np.zeros(condition.sum(), dtype=np.int32)
        new_row = np.asarray(mat.row, dtype=np.int32)[condition]
        new_data = np.asarray(mat.data, dtype=float)[condition]
        this_col = coo_array((new_data, (new_row, new_col)), shape=(mat.shape[0], 1))
        return this_col



    if verbose > 1:
        from time import perf_counter
        t0 = perf_counter()
        print("Starting pre-processing for the SDP solver.")

    if mask_matrices is None:
        mask_matrices = {}
    if known_vars is None:
        known_vars = {}
    if semiknown_vars is None:
        semiknown_vars = {}
    var_objective = {} if objective is None else objective.copy()
    var_inequalities = [] if inequalities is None else deepcopy(inequalities)
    var_equalities   = [] if equalities   is None else deepcopy(equalities)

    Fi = {k: v.asformat('dok', copy=False).astype(float, copy=False)
          for k, v in mask_matrices.items()}
    mat_dim = next(iter(Fi.values())).shape[0] if Fi else 0

    variables = set()
    variables.update(Fi.keys())
    variables.update(var_objective.keys())
    for eq in var_equalities:
        variables.update(eq.keys())
    for ineq in var_inequalities:
        variables.update(ineq.keys())
    variables.difference_update(known_vars.keys())

    if feas_as_optim and mask_matrices:
        if verbose > 0:
            print("Feasibility as optimisation detected: overwriting objective"
                  + " to maximize the smallest eigenvalue of the matrix"
                  + " variable.")
        lam = "lambda"
        while lam in variables:
            lam += "_"
        variables.add(lam)
        Fi[lam] = -1 * eye(mat_dim).tolil()
        var_objective = {lam: 1}

    # Calculate F0, the constant part of the matrix variable.
    F0 = coo_array((mat_dim, mat_dim), dtype=float)
    F0 += sum(known_vars[x] * Fi[x] for x in set(Fi).intersection(known_vars))

    if process_constraints:
        # For the semiknown constraint x_i = a_i * x_j, add to the Fi of x_j
        # the expression a_i*(Fi of x_i).
        for x, (c, x2) in semiknown_vars.items():
            if mask_matrices:
                val = Fi.pop(x, 0)
                Fi[x2] = Fi.get(x2, 0) + c * val

            if var_objective and not feas_as_optim:
                val = var_objective.pop(x, 0)
                var_objective[x2] = var_objective.get(x2, 0) + c * val

            for equality in var_equalities:
                val = equality.pop(x, 0)
                equality[x2] = equality.get(x2, 0) + c * val

            for inequality in var_inequalities:
                val = inequality.pop(x, 0)
                inequality[x2] = inequality.get(x2, 0) + c * val

            variables.remove(x)
            variables.add(x2)

    else:
        # Just add semiknown constraints as equality constraints.
        for x, (c, x2) in semiknown_vars.items():
            var_equalities.append({x: 1, x2: -c})
            variables.add(x2)

    # Calculate c0, the constant part of the var_objective.
    c0 = 0. + float(sum([var_objective[x] * known_vars[x]
                         for x in set(var_objective).intersection(known_vars)])
                    )

    # 'var2index' should be computed after there is no more further modification
    # to 'variables' or any of the constraint or objective dictionaries
    var2index = {x: i for i, x in enumerate(variables)}

    # Calculate the matrices A, C and vectors b, d such that
    # Ax + b >= 0, Cx + d == 0.
    nof_variables = len(variables)
    def constraint_dicts_to_sparse(constraints: List[dict]) -> (coo_array, coo_array):
        nof_constraints = len(constraints)
        A_row = []
        A_col = []
        A_data = []
        b_row = np.arange(nof_constraints, dtype=np.int32)
        b_col = np.zeros(nof_constraints, dtype=np.int32)
        b_data = []
        for i, constraint in enumerate(constraints):
            all_vars_temp = set(constraint)
            free_vars_temp = all_vars_temp.difference(known_vars)
            known_vars_temp = all_vars_temp.intersection(known_vars)
            for x in free_vars_temp:
                A_row.append(i)
                A_col.append(var2index[x])
                A_data.append(constraint[x])
            b_val = 0.
            for x in known_vars_temp:
                x_val = known_vars[x]
                coeff = constraint[x]
                b_val += (coeff * x_val)
            b_data.append(b_val)
        A = coo_array((A_data, (np.array(A_row, dtype=np.int32),
                                np.array(A_col, dtype=np.int32))),
                      shape=(nof_constraints, nof_variables),
                      dtype=float)
        b = coo_array((b_data, (b_row, b_col)),
                      shape=(nof_constraints, 1),
                      dtype=float)
        return A, b
    A, b = constraint_dicts_to_sparse(var_inequalities)
    C, d = constraint_dicts_to_sparse(var_equalities)
    collect()
    if verbose > 1:
        print("Pre-processing took",
              format(perf_counter() - t0, ".4f"),
              "seconds.")
        t0 = perf_counter()
    if verbose > 0:
        print("Building the model...")

    with Model("SDP") as M:
        if solve_dual:
            # Define variables
            if mask_matrices:
                Z = M.variable("Z", Domain.inPSDCone(mat_dim))
            if var_inequalities:
                I = M.variable("I",
                               len(var_inequalities),
                               Domain.greaterThan(0))
                I_reshaped = I.reshape(I.getShape()[0], 1)
                # It seems MOSEK Fusion API does not allow to pick index i
                # of an expression (A^T I)_i, so we do it manually row by row.
                AtI = []  # \sum_j I_j A_ji as i-th entry of AtI
                for var in variables:
                    slice_ = coo_getcol(A, var2index[var])
                    sparse_slice = scipy_to_mosek(slice_)
                    AtI.append(Expr.dot(sparse_slice, I_reshaped))
            if var_equalities:
                E = M.variable("E", len(var_equalities), Domain.unbounded())
                E_reshaped = E.reshape(E.getShape()[0], 1)
                CtI = []  # \sum_j E_j C_ji as i-th entry of CtI
                for var in variables:
                    slice_ = coo_getcol(C, var2index[var])
                    sparse_slice = scipy_to_mosek(slice_)
                    CtI.append(Expr.dot(sparse_slice, E_reshaped))

            # Define and set objective function
            # c0 + Tr Z F0 + I·b + E·d
            obj_mosek = 0.0
            if not feas_as_optim:
                obj_mosek = float(c0)
            if mask_matrices:
                F0_mosek = scipy_to_mosek(F0)
                obj_mosek = Expr.add(obj_mosek, Expr.dot(Z, F0_mosek))
                del F0_mosek
            if var_inequalities:
                b_mosek = scipy_to_mosek(b)
                obj_mosek = Expr.add(obj_mosek, Expr.dot(b_mosek, I_reshaped))
                del b_mosek
            if var_equalities:
                d_mosek = scipy_to_mosek(d)
                obj_mosek = Expr.add(obj_mosek, Expr.dot(d_mosek, E_reshaped))
                del d_mosek

            M.objective(ObjectiveSense.Minimize, obj_mosek)

            # Add constraints
            # ci + Tr Z Fi + \sum_j I_j A_ji + \sum_j E_j C_ji == 0
            ci_constraints = []
            if default_non_negative:
                domain = Domain.lessThan(0)
            else:
                domain = Domain.equalsTo(0)
            for i, x in enumerate(variables):
                lhs = 0.0
                if var_objective and x in set(var_objective
                                              ).difference(known_vars):
                    ci  = float(var_objective[x])
                    lhs += ci
                try:
                    F = Fi.pop(x)
                    lhs = Expr.add(lhs,
                                   Expr.dot(Z,
                                            scipy_to_mosek(F)))
                except KeyError:
                    pass
                if var_inequalities:
                    lhs = Expr.add(lhs, AtI[i])
                    AtI[i] = None
                if var_equalities:
                    lhs = Expr.add(lhs, CtI[i])
                    CtI[i] = None
                ci_constraints.append(M.constraint(f"c{i}", lhs, domain))
        else:
            # Set up the problem in primal formulation

            # Define variables
            if default_non_negative:
                domain = Domain.greaterThan(0)
            else:
                domain = Domain.unbounded()
            x_mosek = M.variable("x", len(variables), domain)

            if var_inequalities:
                b_mosek = scipy_to_mosek(b)
                A_mosek = scipy_to_mosek(A)
                ineq_constraint = M.constraint("Ineq",
                                               Expr.add(Expr.mul(A_mosek,
                                                                 x_mosek),
                                                        b_mosek),
                                               Domain.greaterThan(0))
                del b_mosek, A_mosek

            if var_equalities:
                d_mosek = scipy_to_mosek(d)
                C_mosek = scipy_to_mosek(C)
                eq_constraint = M.constraint("Eq",
                                             Expr.add(Expr.mul(C_mosek,
                                                               x_mosek),
                                                      d_mosek),
                                             Domain.equalsTo(0))
                del d_mosek, C_mosek

            if mask_matrices:
                G = M.variable("G", Domain.inPSDCone(mat_dim))


                # Add matrix constraints
                constraints = np.empty((mat_dim, mat_dim), dtype=object)
                for i in range(mat_dim):
                    for j in range(i, mat_dim):
                        constraints[i, j] = G.index(i, j)
                for i, j in triu_indices(F0):
                    constraints[i, j] = Expr.sub(constraints[i, j], F0[i, j])
                for i, xi in enumerate(variables.intersection(Fi.keys())):
                    for i_, j_ in triu_indices(Fi[xi]):
                        constraints[i_, j_] = \
                            Expr.sub(constraints[i_, j_],
                                    Expr.mul(Fi[xi][i_, j_],
                                              x_mosek.index(var2index[xi])))

                for i in range(mat_dim):
                    for j in range(i, mat_dim):
                        # G(i,j) - F0(i,j) - sum_i xi Fi(i,j) = 0
                        M.constraint(constraints[i, j], Domain.equalsTo(0))


            mosek_obj = c0
            for x in set(var_objective).difference(known_vars):
                ci = float(var_objective[x])
                mosek_obj = Expr.add(mosek_obj,
                                     Expr.mul(ci, x_mosek.index(var2index[x])))

            M.objective(ObjectiveSense.Maximize, mosek_obj)

        collect()
        if verbose > 1:
            print("Model built in",
                  format(perf_counter() - t0, ".4f"),
                  "seconds.")
            M.writeTask("InflationSDPModel.ptf")
            print("Model saved to InflationSDPModel.ptf.")
            t0 = perf_counter()

        # Solve the problem and process the solution
        if verbose > 0:
            print("Solving the model...")
        xmat, ymat, primal, dual = None, None, None, None
        try:
            if verbose > 0:
                M.setLogHandler(stdout)
            if solverparameters:
                for param, val in solverparameters.items():
                    M.setSolverParam(param, val)
            M.acceptedSolutionStatus(AccSolutionStatus.Anything)
            M.solve()
            if solve_dual:
                x_values = {x: float(-ci_constraints[i].dual()[0])
                            for i, x in enumerate(variables)}
                if mask_matrices:
                    ymat = Z.level().reshape([mat_dim, mat_dim])
                    xmat = F0 + sum([x_values[x] * Fi[x]
                                     for x in set(Fi).difference(known_vars)])
            else:
                x_values = dict(zip(variables, x_mosek.level()))
                if mask_matrices:
                    ymat = G.dual().reshape([mat_dim, mat_dim])
                    xmat = G.level().reshape([mat_dim, mat_dim])

            status = M.getProblemStatus()
            if status == ProblemStatus.PrimalAndDualFeasible:
                status_str = "optimal"
                primal     = M.primalObjValue()
                dual       = M.dualObjValue()

            elif status == ProblemStatus.DualInfeasible:
                status_str = "dual_infeas_cer"
            elif status == ProblemStatus.PrimalInfeasible:
                status_str = "primal_infeas_cer"
            elif status == ProblemStatus.Unknown:
                status_str = "unknown"
                code, desc = mosek.Env.getcodedesc(
                    mosek.rescode(int(M.getSolverIntInfo("optimizeResponse"))))
                error_message = f"Termination code {code}: {desc}"
                if verbose > 0:
                    print("The solution status is unknown.\n" + error_message)
                return {"status": status_str, "error": error_message}
            else:
                status_str = "other"
                error_message = ("Another unexpected problem, " +
                                 f"status {status} has been obtained.")
                print(error_message)
                return {"status": status_str, "error": error_message}
        except OptimizeError as e:
            status_str = "other"
            error_message = f"Optimization failed. Error: {e}"
            print(error_message)
            return {"status": status_str, "error": error_message}
        except SolutionError as e:
            status_str = M.getProblemStatus()
            error_message = f"Solution status: {status_str}. Error: {e}"
            print(error_message)
            return {"status": status_str, "error": error_message}
        except Exception as e:
            status_str = "other"
            error_message = f"Unexpected error: {e}"
            print(error_message)
            return {"status": status_str, "error": error_message}

        if status_str in ["optimal", "primal_infeas_cer", "dual_infeas_cer"]:
            certificate = {x: 0 for x in known_vars}

            # c0(P(a...|x...))
            if not feas_as_optim:
                for x in set(var_objective).intersection(known_vars):
                    certificate[x] += var_objective[x]

            # + Tr[Z*F0(P(a...|x...))]=\sum_i*x_{kn_i}(P(a...|x...))*F_{kn_i}
            if mask_matrices:
                for x in set(Fi).intersection(known_vars):
                    support = Fi[x].nonzero()
                    certificate[x] = np.dot(ymat[support], Fi[x][support].toarray().ravel())

            # + I · b
            if var_inequalities:
                Ivalues = I.level() if solve_dual else -ineq_constraint.dual()
                for i, inequality in enumerate(var_inequalities):
                    for x in set(inequality).intersection(known_vars):
                        certificate[x] += Ivalues[i] * inequality[x]

            # + E · d
            if var_equalities:
                Evalues = E.level() if solve_dual else -eq_constraint.dual()
                for i, equality in enumerate(var_equalities):
                    for x in set(equality).intersection(known_vars):
                        certificate[x] += Evalues[i] * equality[x]

            # Clean entries with coefficient zero
            for x in list(certificate.keys()):
                if np.isclose(certificate[x], 0):
                    del certificate[x]

            # For debugging purposes
            if status_str == "optimal" and verbose > 1:
                TOL = 1e-8  # Constraint tolerance
                if var_inequalities:
                    x = A.todense() \
                        @ np.array(list(x_values.values())) \
                        + b.T.todense()[0]
                    if np.any(x < -TOL):
                        print("Warning: Inequality constraints not satisfied" +
                              f" to {TOL} precision.")
                        print("Inequality constraints and their deviation:")
                        print([(ineq, x[i]) for i, (violated, ineq)
                               in enumerate(zip(x < -TOL, var_inequalities))
                               if violated])
                if var_equalities:
                    x = C.todense() \
                        @ np.array(list(x_values.values())) \
                        + d.T.todense()[0]
                    if np.any(np.abs(x) > TOL):
                        print("Warning: Equality constraints not satisfied " +
                              f"to {TOL} precision.")
                        print("Equality constraints and their deviation:")
                        print([(eq, x[i]) for i, (violated, eq)
                               in enumerate(zip(np.abs(x) > TOL,
                                                var_equalities))
                               if violated])

            vars_of_interest = {"primal_value": primal, "dual_value": dual,
                                "status": status_str, "F": xmat, "Z": ymat,
                                "dual_certificate": certificate,
                                "x": x_values}

            return vars_of_interest
        else:
            return {"status": status_str}


def triu_indices(A: coo_array) -> Iterator[tuple[np.int32, np.int32]]:
    """Helper functions to extract the upper triangular (i,j) matrix indices
     of the nonzero elements of a symmetric sparse matrix."""
    A_coo = A.tocoo(copy=False)
    mask = np.logical_and(A_coo.data != 0, A_coo.row <= A_coo.col)
    return zip(A_coo.row[mask].astype(np.int32).flat,
               A_coo.col[mask].astype(np.int32).flat)
